from __future__ import annotations

import math
import torch
import torch.nn as nn
import time
import triton
import triton.language as tl
from tqdm import tqdm
from functools import partial
from typing import Optional, Sequence, Callable, Union, Tuple, Dict

from pyjuice.nodes import CircuitNodes, InputNodes, ProdNodes, SumNodes, foreach, summate, multiply
from pyjuice.layer import Layer, InputLayer, ProdLayer, SumLayer, LayerGroup
from pyjuice.utils.grad_fns import ReverseGrad
from pyjuice.utils import BitSet

from .backend import compile_cum_par_flows_fn, compute_cum_par_flows, cum_par_flows_to_device, \
                     compile_par_update_fn, em_par_update, par_update_to_device, \
                     normalize_parameters, eval_top_down_probs


def _pc_model_backward_hook(grad, pc, inputs, record_cudagraph, apply_cudagraph, propagation_alg, **kwargs):
    grad = grad.permute(1, 0)
    pc.backward(
        inputs = inputs,
        ll_weights = grad / grad.sum() * grad.size(1),
        compute_param_flows = pc._optim_hyperparams["compute_param_flows"], 
        flows_memory = pc._optim_hyperparams["flows_memory"],
        record_cudagraph = record_cudagraph,
        apply_cudagraph = apply_cudagraph,
        propagation_alg = propagation_alg,
        **kwargs
    )

    return None


def layer_iterator(pc, reverse = False, ret_layer_groups = False, ignore_input_layers = False):
    if not reverse:
        if ret_layer_groups:
            if not ignore_input_layers:
                yield pc.input_layer_group

            for layer_group in pc.inner_layer_groups:
                yield layer_group

        else:
            if not ignore_input_layers:
                for layer in pc.input_layer_group:
                    yield layer

            for layer_group in pc.inner_layer_groups:
                for layer in layer_group:
                    yield layer
    else:
        if ret_layer_groups:
            for layer_group in pc.inner_layer_groups[::-1]:
                yield layer_group

            if not ignore_input_layers:
                yield pc.input_layer_group

        else:
            for layer_group in pc.inner_layer_groups[::-1]:
                for layer in layer_group:
                    yield layer

            if not ignore_input_layers:
                for layer in pc.input_layer_group:
                    yield layer


class TensorCircuit(nn.Module):
    """
    A class for compiled PCs. It is a subclass of `torch.nn.Module`.

    :param root_ns: the root node of the PC's DAG
    :type root_ns: CircuitNodes

    :param layer_sparsity_tol: the maximum allowed fraction for added pseudo edges within every layer (better to set to a small number for sparse/block-sparse PCs)
    :type layer_sparsity_tol: float

    :param max_num_partitions: maximum number of partitions in a layer
    :type max_num_partitions: Optional[int]

    :param disable_gpu_compilation: force PyJuice to use CPU compilation
    :type disable_gpu_compilation: bool

    :param force_gpu_compilation: force PyJuice to use GPU compilation
    :type force_gpu_compilation: bool

    :param max_tied_ns_per_parflow_block: how many groups of tied parameters are allowed to share the same flow/gradient accumulator (higher values -> consumes less GPU memory; lower values -> potentially avoid stalls caused by atomic operations)
    :type max_tied_ns_per_parflow_block: int

    :param verbose: Whether to display the progress of the compilation
    :type verbose: bool
    """

    def __init__(self, root_ns: CircuitNodes, layer_sparsity_tol: float = 0.5, 
                 max_num_partitions: Optional[int] = None, disable_gpu_compilation: bool = False, 
                 force_gpu_compilation: bool = False,
                 max_tied_ns_per_parflow_block: int = 8,
                 device: Optional[Union[int,torch.device]] = None,
                 verbose: bool = True) -> None:

        super(TensorCircuit, self).__init__()

        assert isinstance(root_ns, CircuitNodes), "`root_ns` should be an instance of `CircuitNodes`."

        self.root_ns = root_ns
        self.device = torch.device("cpu")

        self.num_vars = self._get_num_vars(self.root_ns)

        self.node_mars = None
        self.element_mars = None
        self.node_flows = None
        self.element_flows = None
        self.param_flows = None
        
        self._init_layers(
            layer_sparsity_tol = layer_sparsity_tol, 
            max_num_partitions = max_num_partitions, 
            disable_gpu_compilation = disable_gpu_compilation, 
            force_gpu_compilation = force_gpu_compilation,
            max_tied_ns_per_parflow_block = max_tied_ns_per_parflow_block,
            device = device,
            verbose = verbose
        )
        
        # Hyperparameters for backward pass
        self._optim_hyperparams = {
            "compute_param_flows": True,
            "flows_memory": 1.0
        }

        # Partial evaluation
        self._fw_partial_eval_enabled = False
        self._bk_partial_eval_enabled = False

        # CudaGraph options
        self._recorded_cuda_graphs = dict()

        # Mode for forward and backward pass
        self.default_propagation_alg = "LL" # Could be "LL", "MPE", or "GeneralLL"
        self.propagation_alg_kwargs = dict()

        # Running parameters
        self._run_params = dict()

        # Cumulative flows
        self._cum_flow = 0.0

    def to(self, device):
        super(TensorCircuit, self).to(device)

        self.input_layer_group.to(device)

        self.device = device

        # For parameter flow accumulation
        self.parflow_fusing_kwargs = cum_par_flows_to_device(self.parflow_fusing_kwargs, device)
        
        # For parameter update
        self.par_update_kwargs = par_update_to_device(self.par_update_kwargs, device)

        return self

    def set_propagation_alg(self, propagation_alg: str, **kwargs):
        if propagation_alg == "LL":
            self.default_propagation_alg = "LL"
            self.propagation_alg_kwargs.clear()
        elif propagation_alg == "MPE":
            self.default_propagation_alg = "MPE"
            self.propagation_alg_kwargs.clear()
        elif propagation_alg == "GeneralLL":
            assert "alpha" in kwargs, "Argument `alpha` should be provided for the `GeneralLL` propagation algorithm."
            self.default_propagation_alg = "GeneralLL"
            self.propagation_alg_kwargs.clear()
            self.propagation_alg_kwargs["alpha"] = kwargs["alpha"]
        else:
            raise NotImplementedError(f"Unknown propagation algorithm {propagation_alg}.")
        
    def forward(self, inputs: torch.Tensor, input_layer_fn: Optional[Union[str,Callable]] = None,
                cache: Optional[dict] = None, return_cache: bool = False, record_cudagraph: bool = False, 
                apply_cudagraph: bool = True, force_use_bf16: bool = False, force_use_fp32: bool = False, 
                propagation_alg: Optional[Union[str,Sequence[str]]] = None, _inner_layers_only: bool = False, 
                _no_buffer_reset: bool = False, **kwargs):
        """
        Forward evaluation of the PC.

        :param inputs: input tensor of size `[B, num_vars]`
        :type inputs: torch.Tensor

        :param input_layer_fn: Custom forward function for input layers; if it is a string, then try to call the corresponding member function of the input layers
        :type input_layer_fn: Optional[Union[str,Callable]]
        """
        
        B = inputs.size(0)

        origin_inputs = inputs
        if input_layer_fn is None:
            assert inputs.dim() == 2

            inputs = inputs.permute(1, 0)

        # Set propagation algorithm
        if propagation_alg is None:
            propagation_alg = self.default_propagation_alg
            kwargs.update(self.propagation_alg_kwargs)
        
        ## Initialize buffers for forward pass ##

        if not _no_buffer_reset:
            self._init_buffer(name = "node_mars", shape = (self.num_nodes, B), set_value = 0.0)
            self._init_buffer(name = "element_mars", shape = (self.num_elements, B), set_value = -torch.inf)

        # Load cached node marginals
        if self._buffer_matches(name = "node_mars", cache = cache):
            self.node_mars[:,:] = cache["node_mars"]

        ## Run forward pass ##

        with torch.no_grad():
            # Input layers
            if not _inner_layers_only:
                for idx, layer in enumerate(self.input_layer_group):
                    if input_layer_fn is None:
                        layer(inputs, self.node_mars, **kwargs)

                    elif isinstance(input_layer_fn, str):
                        assert hasattr(layer, input_layer_fn), f"Custom input function `{input_layer_fn}` not found for layer type {type(layer)}."
                        getattr(layer, input_layer_fn)(inputs, self.node_mars, **kwargs)

                    elif isinstance(input_layer_fn, Callable):
                        input_layer_fn(layer, inputs, self.node_mars, **kwargs)

                    else:
                        raise ValueError(f"Custom input function should be either a `str` or a `Callable`. Found {type(input_layer_fn)} instead.")

            # Inner layers
            def _run_inner_layers():
                for layer_id, layer_group in enumerate(self.inner_layer_groups):
                    if layer_group.is_prod():
                        # Prod layer
                        layer_group(self.node_mars, self.element_mars)

                    elif layer_group.is_sum():
                        # Sum layer
                        layer_group(self.node_mars, self.element_mars, self.params, 
                                    force_use_bf16 = force_use_bf16,
                                    force_use_fp32 = force_use_fp32, 
                                    propagation_alg = propagation_alg if isinstance(propagation_alg, str) else propagation_alg[layer_id], 
                                    **kwargs)

                    else:
                        raise ValueError(f"Unknown layer type {type(layer)}.")

            signature = (0, id(self.node_mars), id(self.element_mars), id(self.params), B)
            if record_cudagraph and signature not in self._recorded_cuda_graphs:
                # Warmup
                s = torch.cuda.Stream()
                s.wait_stream(torch.cuda.current_stream())
                with torch.cuda.stream(s):
                    for _ in range(3):
                        _run_inner_layers()
                torch.cuda.current_stream().wait_stream(s)

                # Capture
                g = torch.cuda.CUDAGraph()
                with torch.cuda.graph(g):
                    _run_inner_layers()

                # Save
                self._recorded_cuda_graphs[signature] = g

            if apply_cudagraph and signature in self._recorded_cuda_graphs:
                g = self._recorded_cuda_graphs[signature]
                g.replay()
            else:
                _run_inner_layers()
                
        lls = self.node_mars[self._root_node_range[0]:self._root_node_range[1],:]
        lls = lls.permute(1, 0)

        ## Create/Update cache if needed ##

        if return_cache:
            if cache is None:
                cache = dict()

            with torch.no_grad():
                cache["node_mars"] = self.node_mars.clone()

        ## Add gradient hook for backward pass ##

        if torch.is_grad_enabled():
            lls.requires_grad = True
            lls.register_hook(
                partial(
                    _pc_model_backward_hook, 
                    pc = self, 
                    inputs = origin_inputs, 
                    record_cudagraph = record_cudagraph, 
                    apply_cudagraph = apply_cudagraph,
                    propagation_alg = propagation_alg,
                    **kwargs
                )
            )

        if return_cache:
            return lls.clone(), cache
        else:
            return lls.clone()

    def backward(self, inputs: Optional[torch.Tensor] = None, 
                 ll_weights: Optional[torch.Tensor] = None,
                 compute_param_flows: bool = True, 
                 flows_memory: float = 1.0,
                 input_layer_fn: Optional[Union[str,Callable]] = None,
                 cache: Optional[dict] = None,
                 sum_layer_pre_backward_callback: Optional[Callable] = None,
                 sum_layer_post_backward_callback: Optional[Callable] = None,
                 return_cache: bool = False,
                 record_cudagraph: bool = False, 
                 apply_cudagraph: bool = True,
                 allow_modify_flows: bool = True,
                 propagation_alg: Union[str,Sequence[str]] = "LL",
                 logspace_flows: bool = False,
                 negate_pflows: bool = False,
                 _inner_layers_only: bool = False,
                 _disable_buffer_init: bool = False,
                 **kwargs):
        """
        Backward evaluation of the PC that computes node flows as well as parameter flows.

        :param inputs: input tensor of size `[B, num_vars]`
        :type inputs: torch.Tensor

        :param ll_weights: weights of the log-likelihoods of size [B] or [num_roots, B]
        :type ll_weights: torch.Tensor
        
        :param input_layer_fn: Custom forward function for input layers; if it is a string, then try to call the corresponding member function of the input layers
        :type input_layer_fn: Optional[Union[str,Callable]]
        """

        self._run_params["allow_modify_flows"] = allow_modify_flows
        self._run_params["propagation_alg"] = propagation_alg
        self._run_params["logspace_flows"] = logspace_flows
        self._run_params["negate_pflows"] = negate_pflows

        assert self.node_mars is not None and self.element_mars is not None, "Should run forward path first."
        if input_layer_fn is None:
            assert inputs.dim() == 2 and inputs.size(1) == self.num_vars
            inputs = inputs.permute(1, 0)

        B = self.node_mars.size(1)

        ## Initialize buffers for backward pass ##

        if not _disable_buffer_init:
            self._init_buffer(name = "node_flows", shape = (self.num_nodes, B), set_value = 0.0 if not logspace_flows else -float("inf"))
            self._init_buffer(name = "element_flows", shape = (self.num_elements, B), set_value = 0.0 if not logspace_flows else -float("inf"))

        # Set root node flows
        def _set_root_node_flows():
            nonlocal ll_weights
            nonlocal logspace_flows
            if ll_weights is None:
                root_flows = 1.0 if not logspace_flows else 0.0
                self.node_flows[self._root_node_range[0]:self._root_node_range[1],:] = root_flows
            else:
                if ll_weights.dim() == 1:
                    ll_weights = ll_weights.unsqueeze(0)

                assert ll_weights.size(0) == self.num_root_nodes

                root_flows = ll_weights if not logspace_flows else ll_weights.log()
                self.node_flows[self._root_node_range[0]:self._root_node_range[1],:] = root_flows

        _set_root_node_flows()

        # Accumulate the total amount of flows added to the PC
        if compute_param_flows:
            if ll_weights is None:
                self._cum_flow += (self._root_node_range[1] - self._root_node_range[0]) * B
            else:
                self._cum_flow += ll_weights.sum().item()

        # Load cached node flows
        if self._buffer_matches(name = "node_flows", cache = cache):
            self.node_flows[:,:] = cache["node_flows"]

        ## Initialize parameter flows ##
        if compute_param_flows:
            self.init_param_flows(flows_memory = flows_memory)

        ## Run backward pass ##

        with torch.no_grad():

            # Inner layers
            def _run_inner_layers():

                # Backward pass for inner layers
                for layer_id in range(len(self.inner_layer_groups) - 1, -1, -1):
                    layer_group = self.inner_layer_groups[layer_id]

                    if layer_group.is_prod():
                        # Prod layer
                        layer_group.backward(self.node_flows, self.element_flows, logspace_flows = logspace_flows)

                    elif layer_group.is_sum():
                        # Sum layer

                        # First recompute the previous product layer
                        self.inner_layer_groups[layer_id-1].forward(self.node_mars, self.element_mars, _for_backward = True)

                        # Execute pre-backward callback
                        layer_group.callback(
                            sum_layer_pre_backward_callback, 
                            node_flows = self.node_flows,
                            element_flows = self.element_flows,
                            node_mars = self.node_mars,
                            element_mars = self.element_mars,
                            params = self.params,
                            param_flows = self.param_flows if hasattr(self, "param_flows") else None
                        )

                        # Backward sum layer
                        layer_group.backward(self.node_flows, self.element_flows, self.node_mars, self.element_mars, self.params, 
                                             param_flows = self.param_flows if compute_param_flows else None,
                                             allow_modify_flows = allow_modify_flows, 
                                             propagation_alg = propagation_alg if isinstance(propagation_alg, str) else propagation_alg[layer_id], 
                                             logspace_flows = logspace_flows, negate_pflows = negate_pflows, **kwargs)

                        # Execute post-backward callback
                        layer_group.callback(
                            sum_layer_post_backward_callback, 
                            node_flows = self.node_flows,
                            element_flows = self.element_flows,
                            node_mars = self.node_mars,
                            element_mars = self.element_mars,
                            params = self.params,
                            param_flows = self.param_flows if hasattr(self, "param_flows") else None
                        )

                    else:
                        raise ValueError(f"Unknown layer type {type(layer)}.")

            signature = (1, id(self.node_flows), id(self.element_flows), id(self.node_mars), id(self.element_mars), id(self.params), id(self.param_flows), B, allow_modify_flows, logspace_flows)
            if record_cudagraph and signature not in self._recorded_cuda_graphs:
                # Warmup
                s = torch.cuda.Stream()
                s.wait_stream(torch.cuda.current_stream())
                with torch.cuda.stream(s):
                    for _ in range(3):
                        self.node_flows[:,:] = 0.0
                        _set_root_node_flows()
                        _run_inner_layers()
                torch.cuda.current_stream().wait_stream(s)

                # Capture
                self.node_flows[:,:] = 0.0
                _set_root_node_flows()
                g = torch.cuda.CUDAGraph()
                with torch.cuda.graph(g):
                    _run_inner_layers()

                # Save
                self._recorded_cuda_graphs[signature] = g

            if apply_cudagraph and signature in self._recorded_cuda_graphs:
                g = self._recorded_cuda_graphs[signature]
                g.replay()
            else:
                _run_inner_layers()

            # Compute backward pass for all input layers
            if not _inner_layers_only:
                for idx, layer in enumerate(self.input_layer_group):
                    if input_layer_fn is None:
                        layer.backward(inputs, self.node_flows, self.node_mars, logspace_flows = logspace_flows, **kwargs)

                    elif isinstance(input_layer_fn, str):
                        assert hasattr(layer, input_layer_fn), f"Custom input function `{input_layer_fn}` not found for layer type {type(layer)}."
                        getattr(layer, input_layer_fn)(inputs, self.node_flows, self.node_mars, logspace_flows = logspace_flows, **kwargs)

                    elif isinstance(input_layer_fn, Callable):
                        input_layer_fn(layer, inputs, self.node_flows, self.node_mars, logspace_flows = logspace_flows, **kwargs)

                    else:
                        raise ValueError(f"Custom input function should be either a `str` or a `Callable`. Found {type(input_layer_fn)} instead.")

        if return_cache:
            if cache is None:
                cache = dict()

            with torch.no_grad():
                cache["node_flows"] = self.node_flows.clone()

            return cache
        else:
            return None

    def forward_ll(self, *args, **kwargs):
        self.forward(*args, propagation_alg = "LL", **kwargs)

    def forward_mpe(self, *args, **kwargs):
        self.forward(*args, propagation_alg = "MPE", **kwargs)

    def forward_general_ll(self, *args, alpha: float = 1.0, **kwargs):
        self.forward(*args, propagation_alg = "GeneralLL", **kwargs)

    def mini_batch_em(self, step_size: float, pseudocount: float = 0.0, keep_zero_params: bool = False,
                      step_size_rescaling: bool = False, use_cudagraph: bool = True):
        """
        Perform an EM parameter update step using the accumulated parameter flows.

        :param step_size: Step size - updated_params <- (1-step_size) * params + step_size * new_params
        :type step_size: float

        :param pseudocount: a pseudo count added to the parameter flows
        :type pseudocount: float

        :param keep_zero_params: if set to `True`, do not add pseudocounts to zero parameters
        :type keep_zero_params: bool

        :param step_size_rescaling: whether to rescale the step size by flows
        :type step_size_rescaling: bool
        """
        assert not step_size_rescaling or self._cum_flow > 0.0, "Please perform a backward pass before calling `mini_batch_em`."
        assert 0.0 < step_size <= 1.0, "`step_size` should be between 0 and 1."

        # Apply step size rescaling according to the mini-batch EM objective derivation
        if step_size_rescaling:
            self.init_param_flows(flows_memory = step_size / self._cum_flow)

            eval_top_down_probs(self, update_pflow = True, scale = (1.0 - step_size), use_cudagraph = use_cudagraph)

            self._cum_flow = 0.0 # Zero out the cumulative flow value
            step_size = 1.0 # We have applied the step size within the parameter flows

        # Update input layers
        for layer in self.input_layer_group:
            layer.mini_batch_em(step_size = step_size, pseudocount = pseudocount)

        # Accumulate parameter flows of tied nodes
        compute_cum_par_flows(self.param_flows, self.parflow_fusing_kwargs)

        # Normalize and update parameters
        em_par_update(self.params, self.param_flows, self.par_update_kwargs, 
                      step_size = step_size, pseudocount = pseudocount,
                      keep_zero_params = keep_zero_params)

    def cumulate_flows(self, inputs: torch.Tensor, params: Optional[torch.Tensor] = None):
        with torch.no_grad():
            self.forward(inputs, params)
            self.backward(inputs = inputs, compute_param_flows = True, flows_memory = 1.0)

    def init_param_flows(self, flows_memory: float = 1.0, batch_size: Optional[int] = None):
        """
        Initialize parameter flows.

        :param flows_memory: the number that the current parameter flows (if any) will be multiplied by; equivalent to zeroling the flows if set to 0
        :type flows_memory: float
        """

        assert 0.0 <= flows_memory <= 1.0, f"`flows_memory` should be in [0.0, 1.0]"

        if batch_size is None:
            pflow_shape = (self.num_param_flows,)
        else:
            pflow_shape = (self.num_param_flows, batch_size)
            
        self._init_buffer(name = "param_flows", shape = pflow_shape)

        if flows_memory != 1.0:
            self.param_flows[:] *= flows_memory

        # For input layers
        for layer in self.input_layer_group:
            layer.init_param_flows(flows_memory = flows_memory)

        return None

    def zero_param_flows(self):
        """
        Zero out parameter flows.
        """
        self.init_param_flows(flows_memory = 0.0)

    def update_parameters(self, clone: bool = True):
        """
        Copy parameters from this `TensorCircuit` to the original `CircuitNodes`.

        :param clone: whether to deepcopy parameters
        :type clone: bool
        """
        params = self.params.detach().cpu()

        for ns in self.root_ns:
            if ns.is_sum() and not ns.is_tied():
                ns.update_parameters(params, clone = clone)

        for layer in self.input_layer_group:
            layer.update_parameters()

        return None

    def update_param_flows(self, clone: bool = True, origin_ns_only: bool = True):
        """
        Copy parameter flows from this `TensorCircuit` to the original `CircuitNodes`.

        :param clone: whether to deepcopy parameters
        :type clone: bool
        """
        param_flows = self.param_flows.detach().cpu()

        for ns in self.root_ns:
            if ns.is_sum() and not ns.is_tied():
                ns.update_param_flows(param_flows, clone = clone, origin_ns_only = origin_ns_only)

    def print_statistics(self):
        """
        Print the statistics of the PC.
        """
        print(f"> Number of nodes: {self.num_nodes}")
        print(f"> Number of edges: {self.num_edges}")
        print(f"> Number of sum parameters: {self.num_sum_params}")

    def get_node_mars(self, ns: CircuitNodes):
        """
        Retrieve the node values of `ns` from the previous forward pass.

        :params ns: the target nodes
        :type ns: CircuitNodes
        """
        assert self.root_ns.contains(ns)
        assert hasattr(self, "node_mars") and self.node_mars is not None
        assert hasattr(self, "element_mars") and self.element_mars is not None

        nsid, neid = ns._output_ind_range

        if ns.is_sum() or ns.is_input():
            return self.node_mars[nsid:neid,:].detach()
        else:
            assert ns.is_prod()

            target_layer = None
            for layer_group in self.inner_layer_groups:
                for layer in layer_group:
                    if layer.is_prod() and ns in layer.nodes:
                        target_layer = layer
                        break

                if target_layer is not None:
                    break

            # Rerun the corresponding product layer to get the node values
            layer(self.node_mars, self.element_mars)

            return self.element_mars[nsid:neid,:].detach()

    def get_node_flows(self, ns: CircuitNodes, **kwargs):
        """
        Retrieve the node flows of `ns` from the previous backward pass.

        :params ns: the target nodes
        :type ns: CircuitNodes
        """
        assert self.root_ns.contains(ns)
        assert hasattr(self, "node_flows") and self.node_flows is not None
        assert hasattr(self, "element_flows") and self.element_flows is not None

        nsid, neid = ns._output_ind_range

        if ns.is_sum() or ns.is_input():
            return self.node_flows[nsid:neid,:].detach()
        else:
            assert ns.is_prod()

            layer_id = None
            for idx, layer_group in enumerate(self.inner_layer_groups):
                for layer in layer_group:
                    if layer.is_prod() and ns in layer.nodes:
                        layer_id = idx
                        break

                if layer_id is not None:
                    break

            # Rerun the corresponding product layer to get the node values
            self.inner_layer_groups[layer_id].forward(self.node_mars, self.element_mars, _for_backward = True)
            self.inner_layer_groups[layer_id+1].backward(
                self.node_flows, self.element_flows, self.node_mars, self.element_mars, self.params, 
                param_flows = None, allow_modify_flows = self._run_params["allow_modify_flows"], 
                propagation_alg = self._run_params["propagation_alg"], 
                logspace_flows = self._run_params["logspace_flows"], 
                negate_pflows = self._run_params["negate_pflows"], **kwargs
            )

            return self.element_flows[nsid:neid,:].detach()

    def layers(self, reverse: bool = False, ret_layer_groups: bool = False, ignore_input_layers: bool = False):
        """
        Returns an iterator of all PC layers.

        :param ret_layer_groups: whether to return `LayerGroup`s instead of `Layer`s
        :type ret_layer_groups: bool
        """
        return layer_iterator(self, reverse = reverse, ret_layer_groups = ret_layer_groups, ignore_input_layers = ignore_input_layers)

    def enable_partial_evaluation(self, scopes: Union[Sequence[BitSet],Sequence[int]], 
                                  forward: bool = False, backward: bool = False, overwrite: bool = False):
        # Create scope2nid cache
        self._create_scope2nid_cache()

        if not overwrite and (forward and self._fw_partial_eval_enabled or backward and self._bk_partial_eval_enabled):
            raise RuntimeError("Partial evaluation already enabled, consider calling `disable_partial_evaluation` first.")

        if isinstance(scopes[0], int):
            scopes = [BitSet.from_array([var]) for var in scopes]

        fw_scopes = scopes if forward else None
        bk_scopes = scopes if backward else None

        # Input layers
        for layer in self.input_layer_group:
            layer.enable_partial_evaluation(fw_scopes = fw_scopes, bk_scopes = bk_scopes)

        # Inner layers
        for layer_group in self.inner_layer_groups:
            layer_group.enable_partial_evaluation(fw_scopes = fw_scopes, bk_scopes = bk_scopes)

        if forward:
            self._fw_partial_eval_enabled = True

        if backward:
            self._bk_partial_eval_enabled = True

    def disable_partial_evaluation(self, forward: bool = True, backward: bool = True):
        # Input layers
        for layer in self.input_layer_group:
            layer.disable_partial_evaluation(forward = forward, backward = backward)

        # Inner layers
        for layer_group in self.inner_layer_groups:
            layer_group.disable_partial_evaluation(forward = forward, backward = backward)

        if forward:
            self._fw_partial_eval_enabled = False

        if backward:
            self._bk_partial_eval_enabled = False

    def _init_buffer(self, name: str, shape: Tuple, set_value: Optional[float] = None, check_device: bool = True):
        flag = False
        if not name in self.__dict__:
            flag = True
        
        tensor = self.__dict__[name]
        if not flag and not isinstance(tensor, torch.Tensor):
            flag = True
        
        if not flag and tensor.dim() != len(shape):
            flag = True

        for i, d in enumerate(shape):
            if not flag and tensor.size(i) != d:
                flag = True
        
        if not flag and check_device and self.device.index is not None and tensor.device != self.device:
            flag = True

        if flag:
            self.__dict__[name] = torch.zeros(shape, device = self.device)

        if set_value is not None:
            if len(shape) == 1:
                self.__dict__[name][:] = set_value
            elif len(shape) == 2:
                self.__dict__[name][:,:] = set_value
            elif len(shape) == 3:
                self.__dict__[name][:,:,:] = set_value
            elif len(shape) == 4:
                self.__dict__[name][:,:,:,:] = set_value
            elif len(shape) == 5:
                self.__dict__[name][:,:,:,:,:] = set_value
            else:
                raise ValueError(f"Too many dimensions ({len(shape)}).")

    def _buffer_matches(self, name: str, cache: Optional[dict], check_device: bool = True):
        if cache is None:
            return False

        assert name in self.__dict__

        tensor = self.__dict__[name]
        
        if name not in cache:
            return False

        if tensor.size() != cache[name].size():
            return False

        if check_device and tensor.device != cache[name].device:
            return False

        return True

    def _get_num_vars(self, ns: CircuitNodes):
        num_vars = 0
        for v in ns.scope:
            if (v + 1) > num_vars:
                num_vars = v + 1
        return num_vars

    def _init_layers(self, layer_sparsity_tol: Optional[float] = None, max_num_partitions: Optional[int] = None,
                     disable_gpu_compilation: bool = False, force_gpu_compilation: bool = False, 
                     max_tied_ns_per_parflow_block: int = 8, verbose: bool = True, device: Optional[Union[str,torch.device]] = None):

        if hasattr(self, "input_layer_group") or hasattr(self, "inner_layer_groups"):
            raise ValueError("Attempting to initialize a TensorCircuit for the second time. " + \
                "Please instead create a new TensorCircuit instance by calling `pc = TensorCircuit(root_ns)`.")

        # Clear hooks/pointers used by previous `TensorCircuit`s
        self.root_ns._clear_tensor_circuit_hooks()

        # TOtal number of variables
        pc_num_vars = len(self.root_ns.scope)

        # Create layers
        depth2nodes, num_layers, max_node_block_size, max_ele_block_size = self._create_node_layers()

        self.input_layer_group = None
        self.inner_layer_groups = []

        self.num_dummy_nodes = max_ele_block_size
        self.num_dummy_eles = max_node_block_size
        self.num_dummy_params = max_node_block_size * max_ele_block_size

        # Nodes include `max_ele_block_size` dummy nodes and all input/sum nodes in the PC
        num_nodes = self.num_dummy_nodes

        # Total number of edges
        num_edges = 0

        # Elements include `max_node_block_size` dummy elements and all product nodes in the PC
        num_elements = self.num_dummy_eles

        # Number of parameters
        num_parameters = self.num_dummy_params

        # Number of parameter flows
        num_param_flows = 0

        # Stores distributed parameter flows
        node2tiednodes = dict()

        if verbose:
            print(f"Compiling {num_layers} TensorCircuit layers...")

        # Select device to use
        if device is None:
            device = torch.cuda.current_device()
        elif isinstance(device, torch.device):
            device = device.index

        with torch.cuda.device(f"cuda:{device}"):
            layer_id = 0
            for depth in tqdm(range(num_layers), disable = not verbose):
                if depth == 0:
                    # Input layer
                    signature2nodes = self._categorize_input_nodes(depth2nodes[0]["input"])
                    input_layer_id = 0
                    input_layers = []
                    for signature, nodes in signature2nodes.items():
                        input_layer = InputLayer(
                            nodes = nodes, cum_nodes = num_nodes,
                            max_tied_ns_per_parflow_block = max_tied_ns_per_parflow_block,
                            pc_num_vars = pc_num_vars
                        )

                        # Special treatment for input layers with the `External` distribution
                        if input_layer.dist_signature == "External":
                            scope = BitSet()
                            for ns in input_layer.nodes:
                                scope |= ns.scope
                            vars = torch.sort(torch.tensor(scope.to_list())).values
                            var_idmapping = torch.zeros([pc_num_vars], dtype = torch.long)
                            var_idmapping[vars] = torch.arange(0, vars.size(0))
                            input_layer.register_buffer("var_idmapping", var_idmapping)

                        input_layers.append(input_layer)
                        
                        input_layer_id += 1
                        num_nodes += input_layer.num_nodes

                    self.input_layer_group = LayerGroup(input_layers)

                else:
                    assert len(depth2nodes[depth]["prod"]) > 0 and len(depth2nodes[depth]["sum"]) > 0, \
                        "Depth {}: (# prod nodes: {}, # sum nodes: {})".format(depth, len(depth2nodes[depth]["prod"]), len(depth2nodes[depth]["sum"]))

                    # Product layer(s)
                    gsize2prod_nodes = dict()
                    for ns in depth2nodes[depth]["prod"]:
                        gsize = ns.block_size
                        if gsize not in gsize2prod_nodes:
                            gsize2prod_nodes[gsize] = []
                        gsize2prod_nodes[gsize].append(ns)
                    
                    layer_num_elements = max_node_block_size
                    prod_layers = []
                    for gsize, nodes in gsize2prod_nodes.items():
                        prod_layer = ProdLayer(
                            nodes = nodes, 
                            global_nid_start = layer_num_elements,
                            layer_sparsity_tol = layer_sparsity_tol,
                            max_num_partitions = max_num_partitions,
                            disable_gpu_compilation = disable_gpu_compilation,
                            force_gpu_compilation = force_gpu_compilation
                        )

                        layer_num_elements += prod_layer.num_nodes
                        num_edges += prod_layer.num_edges

                        prod_layers.append(prod_layer)
                    
                    prod_layer_group = LayerGroup(prod_layers)
                    self.inner_layer_groups.append(prod_layer_group)
                    self.add_module(f"prod_layer_{layer_id}", prod_layer_group)

                    if layer_num_elements > num_elements:
                        num_elements = layer_num_elements

                    # Sum layer(s)
                    gsize2sum_nodes = dict()
                    for ns in depth2nodes[depth]["sum"]:
                        gsize = ns.block_size
                        if gsize not in gsize2sum_nodes:
                            gsize2sum_nodes[gsize] = []
                        gsize2sum_nodes[gsize].append(ns)
                    
                    sum_layers = []
                    for gsize, nodes in gsize2sum_nodes.items():
                        sum_layer = SumLayer(
                            nodes = nodes,
                            global_nid_start = num_nodes, 
                            global_pid_start = num_parameters,
                            global_pfid_start = num_param_flows,
                            node2tiednodes = node2tiednodes,
                            layer_sparsity_tol = layer_sparsity_tol,
                            max_num_partitions = max_num_partitions,
                            max_tied_ns_per_parflow_block = max_tied_ns_per_parflow_block,
                            disable_gpu_compilation = disable_gpu_compilation,
                            force_gpu_compilation = force_gpu_compilation
                        )

                        num_nodes += sum_layer.num_nodes
                        num_edges += sum_layer.num_edges
                        num_parameters += sum_layer.num_parameters
                        num_param_flows += sum_layer.num_param_flows

                        sum_layers.append(sum_layer)

                    sum_layer_group = LayerGroup(sum_layers)
                    self.inner_layer_groups.append(sum_layer_group)
                    self.add_module(f"sum_layer_{layer_id}", sum_layer_group)

                    layer_id += 1

        self.num_nodes = num_nodes
        self.num_edges = num_edges
        self.num_elements = num_elements
        self.num_sum_params = num_parameters
        self.num_param_flows = num_param_flows

        # For parameter flow accumulation
        self.parflow_fusing_kwargs = compile_cum_par_flows_fn(node2tiednodes, MAX_NBLOCKS = 2048, BLOCK_SIZE = 2048)
        
        # For parameter update
        self.par_update_kwargs = compile_par_update_fn(self.root_ns, BLOCK_SIZE = 32)

        # Register root nodes
        self.num_root_nodes = self.root_ns.num_nodes
        self._root_node_range = (self.num_nodes - self.num_root_nodes, self.num_nodes)

        # Initialize parameters
        self._init_parameters()

    def _init_parameters(self, perturbation: float = 4.0, pseudocount: float = 0.0):
        for ns in self.root_ns:
            if not ns.is_tied() and (ns.is_sum() or ns.is_input()) and not ns.has_params():
                ns.init_parameters(perturbation = perturbation, recursive = False)

        params = torch.exp(torch.rand([self.num_sum_params]) * -perturbation)
        params[:self.num_dummy_params] = 0.0

        # Copy initial parameters if provided
        for ns in self.root_ns:
            if ns.is_sum() and not ns.is_tied() and ns.has_params():
                ns.gather_parameters(params)

        self._normalize_parameters(params, pseudocount = pseudocount)
        self.params = nn.Parameter(params)

        # Due to the custom inplace backward pass implementation, we do not track 
        # gradient of PC parameters by PyTorch.
        self.params.requires_grad = False

        # Initialize parameters for input layers
        for idx, layer in enumerate(self.input_layer_group):
            layer._init_parameters(perturbation)

    def _normalize_parameters(self, params, pseudocount: float = 0.0):
        if params is not None:
            normalize_parameters(params, self.par_update_kwargs, pseudocount)

    def _create_node_layers(self):
        depth2nodes = dict()
        nodes2depth = dict()

        num_layers = 1
        max_node_block_size = 0
        max_ele_block_size = 0

        def dfs(ns: CircuitNodes):

            nonlocal num_layers
            nonlocal max_node_block_size
            nonlocal max_ele_block_size

            if ns in nodes2depth:
                return
            if ns.is_input():
                nodes2depth[ns] = 0
                if 0 not in depth2nodes:
                    depth2nodes[0] = {"input": []}
                depth2nodes[0]["input"].append(ns)
            else:
                for cs in ns.chs:
                    dfs(cs)

                depth = max(map(lambda ms: nodes2depth[ms], ns.chs)) + (1 if ns.is_prod() else 0)
                num_layers = max(depth + 1, num_layers)
                nodes2depth[ns] = depth

                if depth not in depth2nodes:
                    depth2nodes[depth] = {"sum": [], "prod": []} # lists for sum and product nodes
                
                if ns.is_sum():
                    for idx, cs in enumerate(ns.chs):
                        cs_depth = nodes2depth[cs]
                        if cs_depth < depth:
                            # TODO: Make the block size be 1
                            pass_sum_ns = summate(
                                cs, num_node_blocks = cs.num_node_blocks, block_size = cs.block_size,
                                edge_ids = torch.arange(0, cs.num_node_blocks)[None,:].repeat(2, 1),
                                params = torch.eye(cs.block_size)[None,:,:].repeat(cs.num_node_blocks, 1, 1)
                            )
                            pass_prod_ns = multiply(pass_sum_ns)
                            ns.chs[idx] = pass_prod_ns

                            depth2nodes[cs_depth]["sum"].append(pass_sum_ns)

                            nodes2depth[pass_sum_ns] = cs_depth
                            nodes2depth[pass_prod_ns] = depth

                    depth2nodes[depth]["sum"].append(ns)

                    if ns.block_size > max_node_block_size:
                        max_node_block_size = ns.block_size
                elif ns.is_prod():
                    if ns.block_size > max_ele_block_size:
                        max_ele_block_size = ns.block_size
                else:
                    raise NotImplementedError(f"Unsupported node type {type(n)}.")

        dfs(self.root_ns)

        pns2layer = dict()
        for layer in range(1, len(depth2nodes)):
            for ns in depth2nodes[layer]["sum"]:
                for cs in ns.chs:
                    if cs.is_prod():
                        if id(cs) in pns2layer:
                            assert pns2layer[id(cs)] == layer, "Disallowed circumstance: a product node requested by sum nodes at different layers."
                        else:
                            depth2nodes[layer]["prod"].append(cs)

                            pns2layer[id(cs)] = layer

        return depth2nodes, num_layers, max_node_block_size, max_ele_block_size

    def _categorize_input_nodes(self, nodes: Sequence[InputNodes]):
        signature2nodes = dict()
        for ns in nodes:
            signature = ns.dist.get_signature()
            if signature not in signature2nodes:
                signature2nodes[signature] = []
            signature2nodes[signature].append(ns)

        return signature2nodes

    def _create_scope2nid_cache(self):
        # Input layers
        for idx, layer in enumerate(self.input_layer_group):
            layer._prepare_scope2nids()

        # Inner layers
        prod_scope_eleids = None
        for layer_group in self.inner_layer_groups:
            if layer_group.is_prod():
                prod_scope_eleids = layer_group._prepare_scope2nids()
            else:
                assert layer_group.is_sum()

                layer_group._prepare_scope2nids(prod_scope_eleids)


def compile(ns: CircuitNodes, layer_sparsity_tol: float = 0.5, 
            max_num_partitions: Optional[int] = None, disable_gpu_compilation: bool = False, 
            force_gpu_compilation: bool = False,
            max_tied_ns_per_parflow_block: int = 8,
            device: Optional[Union[int,torch.device]] = None,
            verbose: bool = True) -> nn.Module:
    """
    Compile a PC represented by a DAG into an equivalent `torch.nn.Module`.

    :param ns: the root node of the PC's DAG
    :type ns: CircuitNodes

    :param layer_sparsity_tol: the maximum allowed fraction for added pseudo edges within every layer (better to set to a small number for sparse/block-sparse PCs)
    :type layer_sparsity_tol: float

    :param max_num_partitions: maximum number of partitions in a layer
    :type max_num_partitions: Optional[int]

    :param disable_gpu_compilation: force PyJuice to use CPU compilation
    :type disable_gpu_compilation: bool

    :param force_gpu_compilation: force PyJuice to use GPU compilation
    :type force_gpu_compilation: bool

    :param max_tied_ns_per_parflow_block: how many groups of tied parameters are allowed to share the same flow/gradient accumulator (higher values -> consumes less GPU memory; lower values -> potentially avoid stalls caused by atomic operations)
    :type max_tied_ns_per_parflow_block: int

    :param device: Which GPU do we use for compilation (the default is `torch.cuda.current_device`)
    :type device: Optional[Union[int,torch.device]]

    :param verbose: Whether to display the progress of the compilation
    :type verbose: bool

    :returns: the compiled PC with type `torch.nn.Module`
    """
    return TensorCircuit(ns, layer_sparsity_tol = layer_sparsity_tol, max_num_partitions = max_num_partitions,
                         disable_gpu_compilation = disable_gpu_compilation, force_gpu_compilation = force_gpu_compilation,
                         max_tied_ns_per_parflow_block = max_tied_ns_per_parflow_block, device = device, verbose = verbose)



import torch
import triton
import triton.language as tl


@triton.jit
def _batched_index_set_kernel(target_tensor_ptr, ids_ptr, source_tensor_ptr, 
                              ids_size, batch_size, BLOCK_SIZE: tl.constexpr):

    pid = tl.program_id(axis = 0)
    block_start = pid * BLOCK_SIZE

    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < ids_size * batch_size

    ids_offsets = offsets // batch_size
    batch_offsets = offsets % batch_size

    n_offsets = tl.load(ids_ptr + ids_offsets, mask = mask, other = 0)
    n_offsets = n_offsets * batch_size + batch_offsets

    source_vals = tl.load(source_tensor_ptr + offsets, mask = mask, other = 0)

    tl.store(target_tensor_ptr + n_offsets, source_vals, mask = mask)


def batched_index_set(target_tensor: torch.Tensor, ids: torch.Tensor, source_tensor: torch.Tensor,
                      BLOCK_SIZE: int = 1024):
    ids_size = ids.numel()
    batch_size = target_tensor.size(-1)

    grid = lambda meta: (triton.cdiv(ids_size * batch_size, meta['BLOCK_SIZE']),)
    _batched_index_set_kernel[grid](
        target_tensor_ptr = target_tensor, 
        ids_ptr = ids, 
        source_tensor_ptr = source_tensor,
        ids_size = ids_size, 
        batch_size = batch_size, 
        BLOCK_SIZE = BLOCK_SIZE
    )


@triton.jit
def _batched_index_cum_kernel(target_tensor_ptr, ids_ptr, source_tensor_ptr, 
                              ids_size, batch_size, BLOCK_SIZE: tl.constexpr):

    pid = tl.program_id(axis = 0)
    block_start = pid * BLOCK_SIZE

    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < ids_size * batch_size

    ids_offsets = offsets // batch_size
    batch_offsets = offsets % batch_size

    n_offsets = tl.load(ids_ptr + ids_offsets, mask = mask, other = 0)
    n_offsets = n_offsets * batch_size + batch_offsets

    source_vals = tl.load(source_tensor_ptr + offsets, mask = mask, other = 0)
    target_vals = tl.load(target_tensor_ptr + n_offsets, mask = mask, other = 0)
    new_vals = target_vals + source_vals

    tl.store(target_tensor_ptr + n_offsets, new_vals, mask = mask)


def batched_index_cum(target_tensor: torch.Tensor, ids: torch.Tensor, source_tensor: torch.Tensor,
                      BLOCK_SIZE: int = 1024):
    ids_size = ids.numel()
    batch_size = target_tensor.size(-1)

    grid = lambda meta: (triton.cdiv(ids_size * batch_size, meta['BLOCK_SIZE']),)
    _batched_index_cum_kernel[grid](
        target_tensor_ptr = target_tensor, 
        ids_ptr = ids, 
        source_tensor_ptr = source_tensor,
        ids_size = ids_size, 
        batch_size = batch_size, 
        BLOCK_SIZE = BLOCK_SIZE
    )


@triton.jit
def _index_cum_kernel(target_tensor_ptr, ids_ptr, source_tensor_ptr, 
                      ids_size, BLOCK_SIZE: tl.constexpr):

    pid = tl.program_id(axis = 0)
    block_start = pid * BLOCK_SIZE

    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < ids_size

    n_offsets = tl.load(ids_ptr + offsets, mask = mask, other = 0)

    source_vals = tl.load(source_tensor_ptr + offsets, mask = mask, other = 0)
    target_vals = tl.load(target_tensor_ptr + n_offsets, mask = mask, other = 0)
    new_vals = target_vals + source_vals

    tl.store(target_tensor_ptr + n_offsets, new_vals, mask = mask)


def index_cum(target_tensor: torch.Tensor, ids: torch.Tensor, source_tensor: torch.Tensor,
                      BLOCK_SIZE: int = 1024):
    ids_size = ids.numel()

    grid = lambda meta: (triton.cdiv(ids_size, meta['BLOCK_SIZE']),)
    _index_cum_kernel[grid](
        target_tensor_ptr = target_tensor, 
        ids_ptr = ids, 
        source_tensor_ptr = source_tensor,
        ids_size = ids_size, 
        BLOCK_SIZE = BLOCK_SIZE
    )


from __future__ import annotations

import torch
import numpy as np
import triton
import triton.language as tl
import random
from typing import Union, Callable, Optional
from functools import partial
from numba import njit

from pyjuice.nodes import CircuitNodes
from pyjuice.model import TensorCircuit


@njit
def _assign_cids_ind_target(ind_target, element_pointers, ind_b, num_samples):
    for i in range(ind_target.shape[0]):
        bid = ind_b[i]
        ind_t = element_pointers[bid]
        ind_target[i] = ind_t * num_samples + bid
        element_pointers[bid] = ind_t + 1


@njit
def _assign_nids_ind_target(ind_target, ind_target_sid, node_pointers, ind_ch_count, ind_b, num_samples):
    nid = 0
    for i in range(ind_target.shape[0]):
        if nid < ind_target_sid.shape[0] - 1 and i >= ind_target_sid[nid+1]:
            nid += 1
        bid = ind_b[nid]
        ind_t = node_pointers[bid]
        ind_target[i] = ind_t * num_samples + bid
        node_pointers[bid] = ind_t + 1


@triton.jit
def sample_sum_layer_kernel(nids, cids, pids, node_mars, element_mars, mparams, node_samples, element_samples, 
                            ind_target, ind_n, ind_b, seed, block_size: tl.constexpr, batch_size: tl.constexpr, 
                            num_edges: tl.constexpr, num_samples: tl.constexpr, num_nblocks: tl.constexpr, BLOCK_S: tl.constexpr, 
                            BLOCK_M: tl.constexpr, M_NUM_BLKS: tl.constexpr, BLOCK_K: tl.constexpr, K_NUM_BLKS: tl.constexpr,
                            conditional: tl.constexpr):
    
    pid_s = tl.program_id(0) # ID of size-`BLOCK_S` batches

    # Sample offsets and mask
    offs_sample = tl.arange(0, BLOCK_S) + pid_s * BLOCK_S
    mask_sample = offs_sample < num_samples

    # Load node and batch ids
    node_sample_id = tl.load(ind_n + offs_sample, mask = mask_sample, other = 0)
    batch_id = tl.load(ind_b + offs_sample, mask = mask_sample, other = 0)
    node_id = tl.load(node_samples + node_sample_id * batch_size + batch_id)

    # Locate node ids in `nids`
    offs_nids = tl.arange(0, BLOCK_M)
    local_nids = tl.zeros([BLOCK_S], dtype = tl.int64) - 1
    local_nid_offs = tl.zeros([BLOCK_S], dtype = tl.int64)
    for i in range(M_NUM_BLKS):
        mask_nids = offs_nids < num_nblocks

        ref_nid = tl.load(nids + offs_nids, mask = mask_nids, other = 0)
        is_match = (node_id[:,None] >= ref_nid[None,:]) & (node_id[:,None] < ref_nid[None,:] + block_size)

        match_local_id = tl.sum(is_match * (offs_nids[None,:] + 1), axis = 1)
        match_local_offset = tl.sum(is_match * (node_id[:,None] - ref_nid[None,:]), axis = 1)

        local_nids = tl.where(match_local_id > 0, match_local_id - 1, local_nids)
        local_nid_offs = tl.where(match_local_id > 0, match_local_offset, local_nid_offs)

        offs_nids += BLOCK_M

    # Update sample mask to filter out inactive ones
    mask_sample = mask_sample & (local_nids >= 0)

    # Sample random probabilities uniform between 0 and 1
    rnd_val = tl.rand(seed, tl.arange(0, BLOCK_S))

    if conditional:
        nmars = tl.load(node_mars + node_id * batch_size + batch_id, mask = mask_sample, other = 0.0) # [Block_B]

    # Calibration loop
    sum_pars = tl.zeros([BLOCK_S], dtype = tl.float32)
    offs_child = tl.arange(0, BLOCK_K)
    mask_child = offs_child < num_edges
    for i in range(K_NUM_BLKS):

        # Load parameters
        param_id = tl.load(pids + local_nids[None,:] * num_edges + offs_child[:,None], mask = (mask_sample[None,:] & mask_child[:,None]), other = 0)
        epars = tl.load(mparams + param_id + local_nid_offs[None,:], mask = (mask_sample[None,:] & mask_child[:,None]), other = 0.0) # [BLOCK_K, BLOCK_B]

        if conditional:
            # In this case, we use `param * cmar / nmar` as the "parameter"
            emars_id = tl.load(cids + local_nids[None,:] * num_edges + offs_child[:,None], mask = (mask_sample[None,:] & mask_child[:,None]), other = 0)
            emars = tl.load(element_mars + emars_id * batch_size + batch_id, mask = (mask_sample[None,:] & mask_child[:,None]), other = 0.0)

            epars = epars * tl.exp(emars - nmars[None,:]) # [BLOCK_K, BLOCK_B]

        sum_pars += tl.sum(epars, axis = 0)

        offs_child += BLOCK_K
        mask_child = offs_child < num_edges

    rnd_val *= sum_pars
    rnd_val -= 1e-8

    # Main loop over blocks of child nodes
    chids = tl.zeros([BLOCK_S], dtype = tl.int64) - 1
    offs_child = tl.arange(0, BLOCK_K)
    mask_child = offs_child < num_edges
    for i in range(K_NUM_BLKS):

        # Load parameters
        param_id = tl.load(pids + local_nids[None,:] * num_edges + offs_child[:,None], mask = (mask_sample[None,:] & mask_child[:,None]), other = 0)
        epars = tl.load(mparams + param_id + local_nid_offs[None,:], mask = (mask_sample[None,:] & mask_child[:,None]), other = 0.0) # [BLOCK_K, BLOCK_B]

        if conditional:
            # In this case, we use `param * cmar / nmar` as the "parameter"
            emars_id = tl.load(cids + local_nids[None,:] * num_edges + offs_child[:,None], mask = (mask_sample[None,:] & mask_child[:,None]), other = 0)
            emars = tl.load(element_mars + emars_id * batch_size + batch_id, mask = (mask_sample[None,:] & mask_child[:,None]), other = 0.0)

            epars = epars * tl.exp(emars - nmars[None,:]) # [BLOCK_K, BLOCK_B]
        
        cum_probs = tl.cumsum(epars, axis = 0) # [BLOCK_K, BLOCK_S]
        local_chids = tl.sum((rnd_val[None,:] >= cum_probs).to(tl.int64), axis = 0) # [BLOCK_S]

        is_overflow = (local_chids == BLOCK_K)
        rnd_val = tl.where(is_overflow, rnd_val - tl.sum(epars, axis = 0), rnd_val)

        chids = tl.where(is_overflow | (chids > -1), chids, local_chids + i * BLOCK_K)

        offs_child += BLOCK_K
        mask_child = offs_child < num_edges

    # Retrieve the global child ids and save them to `element_samples`
    global_chids = tl.load(cids + local_nids * num_edges + chids, mask = mask_sample, other = 0)
    target_id = tl.load(ind_target + offs_sample, mask = mask_sample, other = 0)

    tl.store(element_samples + target_id, global_chids, mask = mask_sample)


def sample_sum_layer(layer, nids, cids, pids, node_mars, element_mars, params, node_samples, element_samples, 
                     ind_target, ind_n, ind_b, block_size, conditional):
    
    num_samples = ind_n.size(0)
    num_nblocks = nids.size(0)
    num_edges = cids.size(1)
    batch_size = node_samples.size(1)
    seed = random.randint(0, 2**31)

    BLOCK_S = min(256, triton.next_power_of_2(num_samples))
    BLOCK_M = min(1024 // BLOCK_S, triton.next_power_of_2(num_nblocks))
    BLOCK_K = min(1024 // BLOCK_S, triton.next_power_of_2(num_edges))

    M_NUM_BLKS = triton.cdiv(num_nblocks, BLOCK_M)
    K_NUM_BLKS = triton.cdiv(num_edges, BLOCK_K)

    grid = (triton.cdiv(num_samples, BLOCK_S),)

    sample_sum_layer_kernel[grid](
        nids, cids, pids, node_mars, element_mars, params, node_samples, element_samples, 
        ind_target, ind_n, ind_b, seed, block_size, batch_size, num_edges, num_samples, num_nblocks, 
        BLOCK_S, BLOCK_M, M_NUM_BLKS, BLOCK_K, K_NUM_BLKS, conditional
    )

    return None


def push_non_neg_ones_to_front(matrix):

    result = torch.full_like(matrix, -1)

    s_mask = (matrix != -1)
    d_mask = torch.sum(s_mask, dim = 0, keepdims = True) > torch.arange(matrix.size(0)).to(matrix.device)[:,None]

    result[d_mask] = matrix[s_mask]
    matrix[:] = result[:]

    return s_mask.long().sum(dim = 0)


@triton.jit
def count_prod_nch_kernel(nids, cids, element_samples, ind_ch_count, ind_nids, ind_nid_offs, ind_mask, ind_n, ind_b, partition_id,
                          block_size: tl.constexpr, num_samples: tl.constexpr, num_nblocks: tl.constexpr, 
                          batch_size: tl.constexpr, num_edges: tl.constexpr, BLOCK_M: tl.constexpr, BLOCK_C: tl.constexpr, 
                          BLOCK_S: tl.constexpr, M_NUM_BLKS: tl.constexpr, C_NUM_BLKS: tl.constexpr):
    
    pid_s = tl.program_id(0) # ID of size-`BLOCK_S` batches

    # Sample offsets and mask
    offs_sample = tl.arange(0, BLOCK_S) + pid_s * BLOCK_S
    mask_sample = offs_sample < num_samples

    # Load node and batch ids
    node_sample_id = tl.load(ind_n + offs_sample, mask = mask_sample, other = 0)
    batch_id = tl.load(ind_b + offs_sample, mask = mask_sample, other = 0)
    ele_id = tl.load(element_samples + node_sample_id * batch_size + batch_id)

    # Locate node ids in `nids`
    offs_nids = tl.arange(0, BLOCK_M)
    local_nids = tl.zeros([BLOCK_S], dtype = tl.int64) - 1
    local_nid_offs = tl.zeros([BLOCK_S], dtype = tl.int64)
    for i in range(M_NUM_BLKS):
        mask_nids = offs_nids < num_nblocks

        ref_nid = tl.load(nids + offs_nids, mask = mask_nids, other = 0)
        is_match = (ele_id[:,None] >= ref_nid[None,:]) & (ele_id[:,None] < ref_nid[None,:] + block_size)

        match_local_id = tl.sum(is_match * (offs_nids[None,:] + 1), axis = 1)
        match_local_offset = tl.sum(is_match * (ele_id[:,None] - ref_nid[None,:]), axis = 1)

        local_nids = tl.where(match_local_id > 0, match_local_id - 1, local_nids)
        local_nid_offs = tl.where(match_local_id > 0, match_local_offset, local_nid_offs)

        offs_nids += BLOCK_M

    # Store `local_nids` and `local_nid_offs` for future reuse
    mask_sample = mask_sample & (local_nids >= 0)
    tl.store(ind_nids + offs_sample, local_nids, mask = mask_sample)
    tl.store(ind_nid_offs + offs_sample, local_nid_offs, mask = mask_sample)
    tl.store(ind_mask + offs_sample, partition_id, mask = mask_sample)

    # Handle triton bug.. (otherwise `local_nids` will be wrong)
    local_nids = tl.load(ind_nids + offs_sample, mask = mask_sample, other = 0)

    # Offset for children
    offs_child = tl.arange(0, BLOCK_C)
    mask_child = offs_child < num_edges

    # Main loop over blocks of child nodes
    ch_count = tl.zeros([BLOCK_S], dtype = tl.int64)
    for i in range(C_NUM_BLKS):

        c_ids = tl.load(cids + local_nids[:,None] * num_edges + offs_child[None,:], mask = (mask_sample[:,None] & mask_child[None,:]), other = 0)
        ch_count += tl.sum((c_ids > 0).to(tl.int64), axis = 1)

        offs_child += BLOCK_C
        mask_child = offs_child < num_edges

    # Store `ch_count`
    tl.store(ind_ch_count + offs_sample, ch_count, mask = mask_sample)


def count_prod_nch(layer, nids, cids, element_samples, ind_ch_count, ind_nids, ind_nid_offs, ind_mask, ind_n, ind_b, block_size, partition_id):

    num_samples = ind_n.size(0)
    num_nblocks = nids.size(0)
    batch_size = element_samples.size(1)
    num_edges = cids.size(1)

    BLOCK_C = min(1024, triton.next_power_of_2(num_edges))
    BLOCK_S = min(1024 // BLOCK_C, triton.next_power_of_2(num_samples))
    BLOCK_M = min(1024 // BLOCK_S, triton.next_power_of_2(num_nblocks))

    M_NUM_BLKS = triton.cdiv(num_nblocks, BLOCK_M)
    C_NUM_BLKS = triton.cdiv(num_edges, BLOCK_C)

    grid = (triton.cdiv(num_samples, BLOCK_S),)

    count_prod_nch_kernel[grid](
        nids, cids, element_samples, ind_ch_count, ind_nids, ind_nid_offs, ind_mask, ind_n, ind_b, partition_id, 
        block_size, num_samples, num_nblocks, batch_size, num_edges, BLOCK_M, BLOCK_C, BLOCK_S, M_NUM_BLKS, C_NUM_BLKS
    )

    return None


@triton.jit
def sample_prod_layer_kernel(nids, cids, node_samples, element_samples, ind_target, ind_target_sid, ind_n, ind_b, 
                             ind_nids, ind_nid_offs, ind_mask, partition_id, block_size: tl.constexpr, 
                             num_samples: tl.constexpr, num_nblocks: tl.constexpr, batch_size: tl.constexpr, num_edges: tl.constexpr,
                             BLOCK_S: tl.constexpr, BLOCK_C: tl.constexpr, C_NUM_BLKS: tl.constexpr):

    pid_s = tl.program_id(0) # ID of size-`BLOCK_S` batches

    # Sample offsets and mask
    offs_sample = tl.arange(0, BLOCK_S) + pid_s * BLOCK_S
    mask_sample = offs_sample < num_samples

    # Load node and batch ids
    node_sample_id = tl.load(ind_n + offs_sample, mask = mask_sample, other = 0)
    batch_id = tl.load(ind_b + offs_sample, mask = mask_sample, other = 0)
    ele_id = tl.load(element_samples + node_sample_id * batch_size + batch_id)

    # Load offsets of `nids` and the node offsets
    local_nids = tl.load(ind_nids + offs_sample, mask = mask_sample, other = 0)
    local_nid_offs = tl.load(ind_nid_offs + offs_sample, mask = mask_sample, other = 0)
    local_partition_id = tl.load(ind_mask + offs_sample, mask = mask_sample, other = 0)

    # Update sample mask
    mask_sample = mask_sample & (local_partition_id == partition_id)

    # Offset for children
    offs_child = tl.arange(0, BLOCK_C)
    mask_child = offs_child < num_edges

    # Main loop over blocks of child nodes
    target_sid = tl.load(ind_target_sid + offs_sample, mask = mask_sample, other = 0)
    for i in range(C_NUM_BLKS):

        c_ids = tl.load(cids + local_nids[:,None] * num_edges + offs_child[None,:], mask = (mask_sample[:,None] & mask_child[None,:]), other = 0)
        target_id = tl.load(ind_target + target_sid[:,None] + offs_child[None,:], mask = (mask_sample[:,None] & mask_child[None,:] & (c_ids > 0)), other = 0)

        tl.store(node_samples + target_id, c_ids + local_nid_offs[:,None], mask = (mask_sample[:,None] & mask_child[None,:] & (c_ids > 0)))

        offs_child += BLOCK_C
        mask_child = offs_child < num_edges


def sample_prod_layer(layer, nids, cids, node_samples, element_samples, ind_target, ind_target_sid, 
                      ind_n, ind_b, ind_nids, ind_nid_offs, ind_mask, block_size, partition_id):
    
    num_samples = ind_n.size(0)
    num_nblocks = nids.size(0)
    num_edges = cids.size(1)
    batch_size = node_samples.size(1)

    BLOCK_C = min(1024, triton.next_power_of_2(num_edges))
    BLOCK_S = min(1024 // BLOCK_C, triton.next_power_of_2(num_samples))

    C_NUM_BLKS = triton.cdiv(num_edges, BLOCK_C)

    grid = (triton.cdiv(num_samples, BLOCK_S),)

    sample_prod_layer_kernel[grid](
        nids, cids, node_samples, element_samples, ind_target, ind_target_sid, ind_n, ind_b, 
        ind_nids, ind_nid_offs, ind_mask, partition_id, block_size, num_samples, 
        num_nblocks, batch_size, num_edges, BLOCK_S, BLOCK_C, C_NUM_BLKS
    )


def sample(pc: TensorCircuit, num_samples: Optional[int] = None, conditional: bool = False, _sample_input_ns: bool = True):
    if not conditional:
        assert num_samples is not None, "`num_samples` should be specified when doing unconditioned sampling."
    else:
        num_samples = pc.node_mars.size(1) # Reuse the batch size

    root_ns = pc.root_ns
    assert root_ns._output_ind_range[1] - root_ns._output_ind_range[0] == 1, "It is ambiguous to sample from multi-head PCs."

    num_nscopes = 0
    num_escopes = 0
    for layer_group in pc.layers(ret_layer_groups = True):
        curr_scopes = 0
        for layer in layer_group:
            curr_scopes += len(layer.scopes)

        if layer_group.is_input() or layer_group.is_sum():
            num_nscopes += curr_scopes
        else:
            assert layer_group.is_prod()
            num_escopes = max(num_escopes, curr_scopes)

    # Stores selected node indices by the sampler
    node_samples = torch.zeros([num_nscopes, num_samples], dtype = torch.long, device = pc.device)
    # Stores selected element indices by the sampler
    element_samples = torch.zeros([num_escopes, num_samples], dtype = torch.long, device = pc.device)
    # Pointers indicating how many elements are used in each column of `element_samples`
    element_pointers = np.zeros([num_samples], dtype = np.int64)

    # Initialize pointers to the root node
    node_samples[:,:] = -1
    node_samples[0,:] = root_ns._output_ind_range[0]

    # Iterate (backward) through layers
    for layer_id in range(len(pc.inner_layer_groups)-1, -1, -1):
        layer_group = pc.inner_layer_groups[layer_id]
        if layer_group.is_sum():
            # Initialize `element_samples` and `element_pointers`
            element_samples[:,:] = -1
            element_pointers[:] = 0

            # Iterate over sum layers in the current layer group
            for layer in layer_group:

                # Gather the indices to be processed
                lsid, leid = layer._layer_nid_range
                ind_n, ind_b = torch.where((node_samples >= lsid) & (node_samples < leid))

                # Pre-compute the target indices in `element_samples`
                ind_target = np.zeros([ind_n.size(0)], dtype = np.int64)
                _assign_cids_ind_target(ind_target, element_pointers, ind_b.detach().cpu().numpy(), num_samples)
                ind_target = torch.from_numpy(ind_target).to(pc.device)

                # In the case of conditional sampling, recompute to get the `element_mars`
                if conditional:
                    pc.inner_layer_groups[layer_id-1](pc.node_mars, pc.element_mars)

                # Sample child indices
                for partition_id in range(layer.num_fw_partitions):
                    nids = layer.partitioned_nids[partition_id]
                    cids = layer.partitioned_cids[partition_id]
                    pids = layer.partitioned_pids[partition_id]
                    
                    sample_sum_layer(layer, nids, cids, pids, pc.node_mars, pc.element_mars, pc.params, 
                                     node_samples, element_samples, ind_target, ind_n, ind_b, 
                                     layer.block_size, conditional)

                # Clear completed nodes
                node_samples[ind_n, ind_b] = -1

        else:
            assert layer_group.is_prod()

            # Iterate over product layers in the current layer group
            for layer in layer_group:
                # Re-align `node_samples` by pushing all values to the front
                node_pointers = push_non_neg_ones_to_front(node_samples)

                # Gather the indices to be processed
                lsid, leid = layer._layer_nid_range
                ind_n, ind_b = torch.where((element_samples >= lsid) & (element_samples < leid))

                # Get the number of children for the selected sample indices
                ind_ch_count = torch.zeros_like(ind_n)
                ind_nids = torch.zeros_like(ind_n)
                ind_nid_offs = torch.zeros_like(ind_n)
                ind_mask = torch.zeros_like(ind_n)
                for partition_id in range(layer.num_fw_partitions):
                    nids = layer.partitioned_nids[partition_id]
                    cids = layer.partitioned_cids[partition_id]

                    count_prod_nch(layer, nids, cids, element_samples, ind_ch_count, ind_nids, 
                                   ind_nid_offs, ind_mask, ind_n, ind_b, layer.block_size, partition_id)

                # Pre-compute the target indices in `node_samples`
                ind_target_sid = np.zeros([ind_n.size(0)], dtype = np.int64)
                ind_target_sid[1:] = ind_ch_count[:-1].cumsum(dim = 0).detach().cpu().numpy()
                ind_target = np.zeros([ind_ch_count.sum()], dtype = np.int64)
                _assign_nids_ind_target(ind_target, ind_target_sid, 
                                        node_pointers.detach().cpu().numpy(), 
                                        ind_ch_count.detach().cpu().numpy(), 
                                        ind_b.detach().cpu().numpy(), num_samples)
                ind_target_sid = torch.from_numpy(ind_target_sid).to(pc.device)
                ind_target = torch.from_numpy(ind_target).to(pc.device)

                # Store child indices
                for partition_id in range(layer.num_fw_partitions):
                    nids = layer.partitioned_nids[partition_id]
                    cids = layer.partitioned_cids[partition_id]

                    sample_prod_layer(layer, nids, cids, node_samples, element_samples, ind_target, ind_target_sid, 
                                      ind_n, ind_b, ind_nids, ind_nid_offs, ind_mask, layer.block_size, partition_id)

    # Create tensor for the samples
    data_dtype = pc.input_layer_group[0].get_data_dtype()
    samples = torch.zeros([pc.num_vars, num_samples], dtype = data_dtype, device = pc.device)

    pc._init_buffer(name = "node_flows", shape = (pc.num_nodes, num_samples), set_value = 0.0)
    ind_n, ind_b = torch.where(node_samples != -1)
    ind_node = node_samples[ind_n, ind_b]
    pc.node_flows[ind_node, ind_b] = 1.0

    if _sample_input_ns:
        for layer in pc.input_layer_group:
            seed = random.randint(0, 2**31)
            layer.sample(samples, pc.node_flows, seed = seed)

        return samples.permute(1, 0).contiguous()
    else:
        # In this case, we do not explicitly sample input nodes
        return node_samples


"""
@Author: Bonan Yan

@Function: Visualization of probabilistic circuits

@Dependency: matplotlib, networkx

@Usage: 
import pyjuice.visualize as juice_vis

# default topology plot: need plt.show() or plt.savefig() following
juice_vis.plot_pc(root_node) 

# display node_num and node_id: need plt.show() or plt.savefig() following
juice_vis.plot_pc(root_node, node_id = True, node_num_label = True) 

"""
import matplotlib.pyplot as plt
import networkx as nx
from pyjuice.io.serialization import serialize_nodes
import numpy as np


def plot_pc(ns, 
           node_id : bool                 = False,
           node_num_label :bool          = False, 
           label_font_size : int         = 8,
           node_num_label_offset : int    = 1,
           node_id_offset : int          = 1) :
    """
    ns: root node
    depend: networkx, matplotlib
    """
    G = nx.DiGraph()
    node_list = serialize_nodes(ns)
    for item in node_list:
        item["num_nodes"] = item["num_node_blocks"] * item["block_size"]
    pos = {}
    nx.set_node_attributes(G, [], "node_type")
    nx.set_node_attributes(G, [], "num_nodes")
    nx.set_node_attributes(G, [], "node_id")

    for i_node in range(len(node_list)):
        G.add_node(i_node)
        
        for j in node_list[i_node]['chs']:
            G.add_edge(j,i_node)
        
        if node_list[i_node]['type']=='Input':
            G.nodes[i_node]['node_type'] = '◯'
        elif node_list[i_node]['type']=='Product':
            G.nodes[i_node]['node_type'] = '×'
        elif node_list[i_node]['type']=='Sum':
            G.nodes[i_node]['node_type'] = '+'
        else:
            G.nodes[i_node]['node_type'] = 'TypeWrong'
        
        G.nodes[i_node]['num_nodes'] = node_list[i_node]['num_nodes']
        G.nodes[i_node]['node_id'] = i_node

    pos = nx.nx_agraph.graphviz_layout(G, prog="dot")
    # pos = nx.nx_agraph.graphviz_layout(G, prog="fdp")
    labels = nx.get_node_attributes(G, 'node_type') 
    options = {
        "with_labels"   : True, 
        "labels"        : labels,
        "node_color"    : "#FFFFFF",
        "font_color"    : "#000000",
        "width"         : 2.0, #note: every cluster of edges is a group of edges, often sparse connection
        "edgecolors"    : "black",
    }
    pos = {node: (x,-y) for (node, (x,y)) in pos.items()}
    nx.draw(G, pos, **options)
    
    if node_num_label:
        attr_num_nodes = nx.get_node_attributes(G, 'num_nodes')
        pos_attrs = {}
        for node, coords in pos.items():
            pos_attrs[node] = (coords[0] + node_num_label_offset, coords[1] + node_num_label_offset)
        nx.draw_networkx_labels(G, 
                                pos_attrs, 
                                font_color = '#E6756F',
                                font_size = label_font_size,
                                labels = attr_num_nodes,
                                verticalalignment = 'top',
                                horizontalalignment = 'left')
    
    if node_id:
        attr_num_nodes = nx.get_node_attributes(G, 'node_id')
        pos_attrs_id = {}
        for node, coords in pos.items():
            pos_attrs_id[node] = (coords[0] + node_id_offset, coords[1] + node_id_offset)
        nx.draw_networkx_labels(G, 
                                pos_attrs_id, 
                                font_color = '#7F7DB5',
                                font_weight='bold',
                                font_size = label_font_size,
                                labels = attr_num_nodes,
                                verticalalignment = 'bottom',
                                horizontalalignment = 'left')
    return


def plot_tensor_node_connection(ns, node_id : int = 0):
    G = nx.DiGraph()
    node_list = serialize_nodes(ns)
    for item in node_list:
        item["num_nodes"] = item["num_node_blocks"] * item["block_size"]
    node_target = node_list[node_id]
    pos = {}
    
    if node_target['type']=='Input':
        print(f"\n>>The target node {node_id} is an Input node, it has {node_target['num_nodes']} nodes, & no connection among them.<<\n")
        return
    
    elif node_target['type']=='Product':
        print(f"\n>>The target node {node_id} is a Product node")
        print(f">> Num_nodes: {node_target['num_nodes']}")
        print(f">> Child num_nodes: {node_list[node_target['chs'][0]]['num_nodes']}\n")
        return np.array(node_target['edge_ids'])
        
    elif node_target['type']=='Sum':
        print(f'The target node [{node_id}] is a Sum node...')
        # add nodes
        sum_node_list = []
        prod_node_list = []
        node_target_prd_chs = node_target['chs'][0] # integer
        adjacency_manual = np.zeros((node_target['num_nodes'], node_list[node_target_prd_chs]['num_nodes']))
        # left, sum: 
        for i in range(node_target['num_nodes']):
            G.add_node(f's{i}')
            sum_node_list.append(f's{i}')
        # right, product:
        for i in range(node_list[node_target_prd_chs]['num_nodes']):
            G.add_node(f'p{i}')
            prod_node_list.append(f'p{i}')
        
        # add edges
        _,c = node_target['edge_ids'].shape
        for i in range(c):
            G.add_edge(f"s{node_target['edge_ids'][0,i]}",f"p{node_target['edge_ids'][1,i]}") 
            adjacency_manual[node_target['edge_ids'][0,i],node_target['edge_ids'][1,i]] = 1
        
        # graph_plot
        # print('>>[Progress]: calculating graph layout...<<')
        # # pos = nx.nx_agraph.graphviz_layout(G, prog="dot")
        # pos = nx.circular_layout(G)
        # print('>>[Progress]: calculated done!<<')
        # options = {
        #     'node_size': 5,
        # }
        # print('>>[Progress]: drawing...<<')
        # nx.draw_networkx_nodes(G, pos, nodelist=sum_node_list, node_color="#A5CE9D", **options) # green is sum nodes
        # nx.draw_networkx_nodes(G, pos, nodelist=prod_node_list, node_color="#A2C8DD", **options) # blue is product nodes
        # nx.draw_networkx_edges(G, pos, width=0.1)
        
        # Adjacency matrix heatmap plot
        fig, ax = plt.subplots()
        ax.spy(adjacency_manual)
        ax.set_aspect('auto')
        ax.set_xlabel('product nodes id')
        ax.set_ylabel('sum nodes id')
        ax.set_title(f"connection adjacency matrix of sum node {node_id} to its child")
        return adjacency_manual # return adjacency matrix
    else:
        print('--Plot Node Type Wrong!--')
        return False
            


import torch
import numpy as np
from numba import njit
import math

from typing import Union, Optional


@njit()
def _partition_nodes_dp_simple_compiled(node_n_edges, dp, backtrace, max_num_partitions, target_overhead):
    num_nodes = node_n_edges.shape[0]

    # Init
    for i in range(num_nodes):
        dp[i,1] = node_n_edges[i] * (i + 1)

    # Main DP
    target_n_partition = max_num_partitions
    for n_partition in range(2, max_num_partitions + 1):
        dp[0,n_partition] = node_n_edges[0]
        backtrace[0,n_partition] = 0
        for i in range(1, num_nodes):
            min_overhead = 2 ** 31 - 1
            best_idx = -1
            for j in range(i):
                curr_overhead = dp[j,n_partition-1] + node_n_edges[i] * (i - j)
                if curr_overhead < min_overhead:
                    min_overhead = curr_overhead
                    best_idx = j

            dp[i,n_partition] = min_overhead
            backtrace[i,n_partition] = best_idx

        if dp[-1,n_partition] < target_overhead:
            target_n_partition = n_partition
            break

    overhead = dp[-1,target_n_partition]

    return overhead, target_n_partition

@njit
def _backtrace_fn(partitions, backtrace, target_n_partition, num_nodes):
    i = num_nodes - 1
    for n in range(target_n_partition, 0, -1):
        partitions[n-1] = i
        i = backtrace[i,target_n_partition]


def _partition_nodes_dp_simple(node_n_edges: np.ndarray, max_num_partitions: int, target_overhead: Optional[int]):

    dp = np.zeros([node_n_edges.shape[0], max_num_partitions + 1], dtype = np.int64)
    backtrace = np.zeros([node_n_edges.shape[0], max_num_partitions + 1], dtype = np.int64)

    overhead, target_n_partition = _partition_nodes_dp_simple_compiled(
        np.ascontiguousarray(node_n_edges), 
        np.ascontiguousarray(dp), 
        np.ascontiguousarray(backtrace),
        max_num_partitions,
        target_overhead = 0 if target_overhead is None else target_overhead
    )

    # Backtrace
    partitions = np.zeros([target_n_partition], dtype = np.int64)
    num_nodes = node_n_edges.shape[0]
    _backtrace_fn(partitions, backtrace, target_n_partition, num_nodes)

    return np.unique(node_n_edges[partitions]), overhead


@njit
def _coalesce_fn(unique_vals, counts, keep_flag, unique_counts, tol_range):
    num_vals = unique_vals.shape[0]

    curr_keep_val = 0
    curr_count = 0
    curr_idx = -1
    for i in range(num_vals):
        if curr_idx == -1:
            curr_idx = i
            curr_keep_val = unique_vals[i]
            curr_count = counts[i]
        elif unique_vals[i] - curr_keep_val <= tol_range:
            curr_count += counts[i]
        else:
            keep_flag[i-1] = True
            unique_counts[i-1] = curr_count

            curr_idx = i
            curr_keep_val = unique_vals[i]
            curr_count = counts[i]

    if curr_idx != -1:
        keep_flag[num_vals-1] = True
        unique_counts[num_vals-1] = curr_count


def _coalesce(vals, tol_range = "auto"):
    unique_vals, counts = np.unique(vals, return_counts = True)
    keep_flag = np.zeros([unique_vals.shape[0]], dtype = bool)
    unique_counts = np.zeros([unique_vals.shape[0]], dtype = np.int64)

    if tol_range == "auto":
        tol_range = int(math.floor((unique_vals.max() - unique_vals.min()) / 1000))

    _coalesce_fn(
        np.ascontiguousarray(unique_vals), 
        np.ascontiguousarray(counts), 
        keep_flag, 
        unique_counts, 
        tol_range
    )

    return unique_vals[keep_flag], unique_counts[keep_flag]


@njit()
def _weighted_partition_nodes_dp_simple_compiled(node_n_edges, cum_counts, dp, backtrace, max_num_partitions, target_overhead):
    num_nodes = node_n_edges.shape[0]

    # Init
    for i in range(num_nodes):
        dp[i,1] = node_n_edges[i] * cum_counts[i]

    if dp[-1,1] <= target_overhead:
        return dp[-1, 1], 1

    # Main DP
    target_n_partition = max_num_partitions
    for n_partition in range(2, max_num_partitions + 1):
        dp[0,n_partition] = node_n_edges[0] * cum_counts[0]
        backtrace[0,n_partition] = 0
        for i in range(1, num_nodes):
            min_overhead = 2 ** 31 - 1
            best_idx = -1
            for j in range(i):
                curr_overhead = dp[j,n_partition-1] + node_n_edges[i] * (cum_counts[i] - cum_counts[j])
                if curr_overhead < min_overhead:
                    min_overhead = curr_overhead
                    best_idx = j

            dp[i,n_partition] = min_overhead
            backtrace[i,n_partition] = best_idx

        if dp[-1,n_partition] <= target_overhead:
            target_n_partition = n_partition
            break

    overhead = dp[-1,target_n_partition]

    return overhead, target_n_partition


def _weighted_partition_nodes_dp_simple(node_n_edges: np.ndarray, counts: np.ndarray, max_num_partitions: int, 
                                        target_overhead: Optional[int]):

    cum_counts = np.cumsum(counts)

    dp = np.zeros([node_n_edges.shape[0], max_num_partitions + 1], dtype = np.int64)
    backtrace = np.zeros([node_n_edges.shape[0], max_num_partitions + 1], dtype = np.int64)

    overhead, target_n_partition = _weighted_partition_nodes_dp_simple_compiled(
        np.ascontiguousarray(node_n_edges),
        np.ascontiguousarray(cum_counts),
        np.ascontiguousarray(dp), 
        np.ascontiguousarray(backtrace),
        max_num_partitions,
        target_overhead = 0 if target_overhead is None else target_overhead
    )

    # Backtrace
    partitions = np.zeros([target_n_partition], dtype = np.int64)
    num_nodes = node_n_edges.shape[0]
    _backtrace_fn(partitions, backtrace, target_n_partition, num_nodes)

    return np.unique(node_n_edges[partitions]), overhead


def partition_nodes_by_n_edges(node_n_edges: Union[np.ndarray, torch.Tensor], 
                               max_num_partitions: Optional[int] = None, 
                               sparsity_tolerance: Optional[float] = None,
                               algorithm: str = "dp_with_coalesce"):

    if sparsity_tolerance is not None and sparsity_tolerance < 1e-6:
        sparsity_tolerance = None
        max_num_partitions = 1

    if sparsity_tolerance is not None:
        assert sparsity_tolerance > 1e-6 and sparsity_tolerance <= 1.0
        if max_num_partitions is None:
            max_num_partitions = max(min(torch.unique(node_n_edges).shape[0], 16), 1)
    elif max_num_partitions is None:
        max_num_partitions = 1
    else:
        assert max_num_partitions >= 1, "Should provide at least 1 partition."

    if isinstance(node_n_edges, torch.Tensor):
        node_n_edges = node_n_edges.detach().cpu().numpy()

    target_overhead = None if sparsity_tolerance is None else int(math.ceil(node_n_edges.sum() * (1.0 + sparsity_tolerance)))

    if max_num_partitions == 1:
        partitions = np.zeros([1], dtype = np.int64)
        partitions[0] = np.max(node_n_edges)
        return torch.from_numpy(partitions)

    # Sort in non-descending order
    node_n_edges = np.sort(node_n_edges)

    if node_n_edges[0] == 0:
        num_zeros = (node_n_edges == 0).sum()
        node_n_edges = node_n_edges[num_zeros:]

    if algorithm == "dp_simple":
        block_sizes, overhead = _partition_nodes_dp_simple(node_n_edges, max_num_partitions, target_overhead)

    elif algorithm == "dp_with_coalesce":
        unique_n_edges, counts = _coalesce(node_n_edges, tol_range = "auto")
        block_sizes, overhead = _weighted_partition_nodes_dp_simple(unique_n_edges, counts, max_num_partitions, target_overhead)

    else:
        raise ValueError(f"Unknown algorithm {algorithm} for `partition_nodes_by_n_edges`.")

    if isinstance(block_sizes, np.ndarray):
        block_sizes = torch.from_numpy(block_sizes)

    return torch.sort(block_sizes).values


from __future__ import annotations

import torch
import triton
import triton.language as tl
import math

from typing import Optional

from pyjuice.nodes import CircuitNodes, InputNodes, ProdNodes, SumNodes
from pyjuice.nodes.distributions import Categorical
from pyjuice.utils.kernel_launcher import FastJITFunction


@triton.jit
def _pairwise_count_kernel(data1_ptr, data2_ptr, pairwise_count_ptr, num_samples: tl.constexpr,
                           n_cls1: tl.constexpr, n_cls2: tl.constexpr, BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(axis = 0)
    block_start = pid * BLOCK_SIZE

    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < num_samples

    cid1 = tl.load(data1_ptr + offsets, mask = mask, other = 0)
    cid2 = tl.load(data2_ptr + offsets, mask = mask, other = 0)
    cid = cid1 * n_cls2 + cid2

    tl.atomic_add(pairwise_count_ptr + cid, 1, mask = mask)


def get_pairwise_count(data1: torch.Tensor, data2: torch.Tensor, n_cls1: int, n_cls2: int, 
                       device: Optional[torch.device] = None, BLOCK_SIZE = 2048):
    assert data1.min() >= 0 and data1.max() < n_cls1, f"Value range of `data1` exceeds limit: [Min: {data1.min().item()}, Max: {data1.max().item()}]."
    assert data2.min() >= 0 and data2.max() < n_cls2, f"Value range of `data2` exceeds limit: [Min: {data2.min().item()}, Max: {data2.max().item()}]."
    assert data1.size(0) == data2.size(0), "`data1` and `data2` must have the same number of examples."

    if device is not None:
        data1 = data1.to(device)
        data2 = data2.to(device)

    if data1.is_cuda:

        data1 = data1.long()
        data2 = data2.long()

        num_samples = data1.size(0)
        pairwise_count = torch.zeros([n_cls1, n_cls2], dtype = torch.float32, device = data1.device)

        grid = lambda meta: (triton.cdiv(num_samples, meta['BLOCK_SIZE']),)

        _pairwise_count_kernel[grid](
            data1_ptr = data1, 
            data2_ptr = data2,
            pairwise_count_ptr = pairwise_count,
            num_samples = num_samples,
            n_cls1 = n_cls1,
            n_cls2 = n_cls2,
            BLOCK_SIZE = BLOCK_SIZE
        )

    else:
        pairwise_count = torch.bincount(data1 * n_cls2 + data2, minlength = n_cls1 * n_cls2)
        pairwise_count = pairwise_count.reshape(n_cls1, n_cls2)

    return pairwise_count


def lvd_for_input_nodes(lvdistiller, ns, lv_dataset: torch.Tensor, obs_dataset: torch.Tensor):
    if isinstance(ns.dist, Categorical):
        pairwise_count = get_pairwise_count(lv_dataset, obs_dataset, ns.num_nodes, ns.dist.num_cats)
        pairwise_count = pairwise_count + lvdistiller.pseudocount
        params = pairwise_count / (pairwise_count.sum(dim = 1, keepdim = True) + 1e-8)
        ns.set_params(params)
    else:
        raise NotImplementedError(f"Input LVD function not implemented for distribution type {type(ns.dist)}.")


def lvd_by_counting(lvdistiller, ns: CircuitNodes):
    lv_dataset_id = lvdistiller.ns2lv_dataset_id[ns]
    if ns.is_prod():
        for cs in ns.chs:
            if cs not in lvdistiller.ns2lv_dataset_id:
                lvdistiller.ns2lv_dataset_id[cs] = lv_dataset_id

    # Get candidate LVD nodes
    nodes_for_lvd = []
    if ns.is_sum():
        if any([cs in lvdistiller.ns2lv_dataset_id for cs in ns.chs]):
            nodes_for_lvd.append(ns)
    elif ns.is_prod():
        for cs in ns.chs:
            if any([ccs in lvdistiller.ns2lv_dataset_id for ccs in cs.chs]):
                nodes_for_lvd.append(cs)
            elif cs.is_input() and cs in lvdistiller.ns2lv_dataset_id and cs in lvdistiller.ns2obs_dataset_id:
                nodes_for_lvd.append(cs)
    elif ns.is_input() and ns in lvdistiller.ns2lv_dataset_id and ns in lvdistiller.ns2obs_dataset_id:
        nodes_for_lvd.append(ns)

    # Run LVD by counting
    for ns in nodes_for_lvd:
        if ns.is_input():
            ns_lv_dataset = lvdistiller.lv_datasets[lvdistiller.ns2lv_dataset_id[ns]]
            ns_obs_dataset = lvdistiller.obs_datasets[lvdistiller.ns2obs_dataset_id[ns]]
            lvd_for_input_nodes(lvdistiller, ns, ns_lv_dataset, ns_obs_dataset)
        else:
            assert ns.is_sum(), "Product nodes cannot apply LVD."
            
            ns_dataset = lvdistiller.lv_datasets[lvdistiller.ns2lv_dataset_id[ns]]
            num_ch_nodes = sum([cs.num_nodes for cs in ns.chs])
            edge_params = torch.empty([ns.num_nodes, num_ch_nodes], dtype = torch.float32, device = ns_dataset.device)
            sid = 0
            for cs in ns.chs:
                eid = sid + cs.num_nodes
                if cs in lvdistiller.ns2lv_dataset_id:
                    cs_dataset = lvdistiller.lv_datasets[lvdistiller.ns2lv_dataset_id[cs]]
                    pairwise_count = get_pairwise_count(ns_dataset, cs_dataset, ns.num_nodes, cs.num_nodes)
                else:
                    # Randomly initialize parameters
                    pairwise_count = torch.exp(-torch.rand([ns.num_nodes, cs.num_nodes]) * 2.0)

                edge_params[:,sid:eid] = pairwise_count

                sid = eid

            # Pruning
            if "prune_frac" in lvdistiller.kwargs and lvdistiller.kwargs["prune_frac"] != 0.0:
                assert 0.0 <= lvdistiller.kwargs["prune_frac"] < 1.0
                edge_mask = ns._get_edges_as_mask()
                num_edges_to_keep = int(math.ceil(ns.num_edges() * (1.0 - lvdistiller.kwargs["prune_frac"])))
                edge_params[~edge_mask] = 0
                _, indices = torch.topk(edge_params.flatten(), k = num_edges_to_keep)
                edge_mask = torch.zeros([ns.num_nodes * ns.num_ch_nodes], dtype = torch.bool)
                edge_mask[indices] = True
                edge_mask = edge_mask.reshape(ns.num_nodes, ns.num_ch_nodes)

                # Make sure that every node has at least one child
                dnids = torch.where(edge_mask.sum(dim = 1) == 0)
                amaxes = torch.argmax(edge_params[dnids,:], dim = 1)
                edge_mask[dnids,amaxes] = True

                # Update edges
                ns._construct_edges(edge_mask)

            edge_params /= edge_params.sum(dim = 1, keepdim = True) + 1e-8
            edge_params = edge_params[:,:,None,None]

            ns.set_params(edge_params, pseudocount = lvdistiller.pseudocount)


import pyjuice.graph
import pyjuice.nodes
import pyjuice.distributions
import pyjuice.layer
import pyjuice.structures
import pyjuice.optim
import pyjuice.transformations
import pyjuice.queries
import pyjuice.io
import pyjuice.visualize

# TensorCircuit
from pyjuice.model import compile, TensorCircuit

# TensorCircuit layers
from pyjuice.layer import InputLayer, ProdLayer, SumLayer

# Construction methods
from pyjuice.nodes import multiply, summate, inputs, set_block_size, structural_properties

# Distributions
from pyjuice.nodes import distributions

# LVD
from pyjuice.nodes.methods.lvd import LVDistiller

# Commonly-used transformations
from pyjuice.transformations import merge, blockify, unblockify, deepcopy

# IO
from pyjuice.io import load, save



from __future__ import annotations

import torch
from functools import partial
from typing import Callable, Optional, Dict

from pyjuice.nodes import CircuitNodes, InputNodes, ProdNodes, SumNodes
from pyjuice.graph import RegionGraph, PartitionNode, InnerRegionNode, InputRegionNode


def merge_sum_nodes(ns1: SumNodes, ns2: SumNodes, *args) -> SumNodes:
    all_ns = [ns1, ns2, *args]
    for idx, ns in enumerate(all_ns):
        assert ns1.scope == ns.scope, "Sum nodes to be merged should have the same scope."
        assert ns1.block_size == ns.block_size, "To-be-merged sum nodes must have the same block size."
        if not isinstance(ns, SumNodes):
            edge_ids = torch.arange(0, ns.num_node_blocks).unsqueeze(0).repeat(2, 1)
            block_size = ns.block_size
            params = torch.eye(ns.block_size).unsqueeze(0).repeat(ns.num_node_blocks, 1, 1)
            new_ns = SumNodes(ns.num_node_blocks, [ns], edge_ids, params = params, block_size = block_size)
            all_ns[idx] = new_ns

    sum_edge_ids = []
    sum_chs = []
    cs2start_id = dict()
    ns_start_id = 0
    global_cs_start_id = 0
    ch_block_size = None
    for ns in all_ns:
        ns_end_id = ns_start_id + ns.num_node_blocks
        curr_cs_sid = 0
        edge_ids = ns.edge_ids.clone()
        for cs in ns.chs:
            if ch_block_size is None:
                ch_block_size = cs.block_size
            else:
                assert ch_block_size == cs.block_size, "Children must have the same block size."

            curr_cs_eid = curr_cs_sid + cs.num_node_blocks
            if cs in cs2start_id:
                cs_start_id = cs2start_id[cs]
            else:
                cs_start_id = global_cs_start_id

            filter = (ns.edge_ids[1,:] >= curr_cs_sid) & (ns.edge_ids[1,:] < curr_cs_eid)
            edge_ids[1,filter] += cs_start_id - curr_cs_sid

            curr_cs_sid = curr_cs_eid
            if cs not in cs2start_id:
                cs2start_id[cs] = global_cs_start_id
                global_cs_start_id += cs.num_node_blocks
                sum_chs.append(cs)

        edge_ids[0,:] += ns_start_id
        sum_edge_ids.append(edge_ids)
        
        ns_start_id = ns_end_id

    num_node_blocks = ns_start_id
    edge_ids = torch.cat(sum_edge_ids, dim = 1)
    if all([hasattr(ns, "_params") and ns._params is not None for ns in all_ns]):
        params = torch.cat([ns._params for ns in all_ns], dim = 0)
    else:
        params = None
    
    return SumNodes(num_node_blocks, sum_chs, edge_ids, params = params, block_size = ns1.block_size)


def merge_prod_nodes(ns1: ProdNodes, ns2: ProdNodes, *args) -> ProdNodes:
    all_ns = [ns1, ns2, *args]
    ch_scopes = [cs.scope for cs in ns1.chs]
    num_scopes = len(ch_scopes)
    for ns in all_ns:
        assert isinstance(ns, ProdNodes), "Inputs should all be ProdNodes."
        assert ns1.scope == ns.scope, "Product nodes to be merged should have the same scope."
        assert ns1.block_size == ns.block_size, "To-be-merged product nodes must have the same block size."
        for cs, scope in zip(ns.chs, ch_scopes):
            assert cs.scope == scope

    cs2start_id = dict()
    sum_chs = [[] for _ in range(num_scopes)]
    global_start_ids = [0 for _ in range(num_scopes)]
    ch_block_size = None
    for ns in all_ns:
        for scope_id in range(num_scopes):
            cs = ns.chs[scope_id]

            if ch_block_size is None:
                ch_block_size = cs.block_size
            else:
                assert ch_block_size == cs.block_size, "Children must have the same block size."

            if cs not in cs2start_id:
                cs2start_id[cs] = global_start_ids[scope_id]
                global_start_ids[scope_id] += cs.num_node_blocks
                sum_chs[scope_id].append(cs)

    new_sum_chs = []
    for scope_id in range(num_scopes):
        sum_ns = sum_chs[scope_id]
        if len(sum_ns) == 1:
            new_sum_chs.append(sum_ns[0])
        else:
            new_sum_chs.append(merge_sum_nodes(*sum_ns))

    prod_edge_ids = []
    use_sparse_mode = any([ns.is_sparse() for ns in all_ns])
    for ns in all_ns:
        edge_ids = ns.edge_ids.clone()
        if use_sparse_mode and ns.is_block_sparse():
            edge_ids = (edge_ids[:,None,:].repeat(1, ns.block_size, 1) * ns.block_size + torch.arange(0, ns.block_size)[None,:,None]).flatten(0, 1)
        for scope_id in range(num_scopes):
            cs = ns.chs[scope_id]
            edge_ids[:,scope_id] += cs2start_id[cs]

        prod_edge_ids.append(edge_ids)

    edge_ids = torch.cat(prod_edge_ids, dim = 0)
    num_node_blocks = edge_ids.size(0)

    return ProdNodes(num_node_blocks, new_sum_chs, edge_ids, block_size = ns1.block_size)


def merge_by_region_node(root_ns: CircuitNodes) -> CircuitNodes:

    rg2nodes = dict()
    rgs_list = list()
    rgs_set = set()
    for ns in root_ns:
        rg = ns.region_node
        rg_hash = hash(rg)
        scope = ns.scope
        if rg_hash in rgs_set:
            rg2nodes[rg_hash].append(ns)
        else:
            rg2nodes[rg_hash] = [ns]

            rgs_list.append(rg)
            rgs_set.add(rg_hash)

    ns_old2new = dict()
    for rg in rgs_list:
        rg_hash = hash(rg)
        if isinstance(rg, InputRegionNode):
            for ns in rg2nodes[rg_hash]:
                ns_old2new[ns] = (ns, (0, ns.num_node_blocks))
        elif isinstance(rg, PartitionNode):
            prod_ns = []
            for ns in rg2nodes[rg_hash]:
                chs = []
                edge_ids = ns.edge_ids.clone()
                for scope_id, cs in enumerate(ns.chs):
                    new_cs, (sid, eid) = ns_old2new[cs]
                    edge_ids[:,scope_id] += sid
                    chs.append(new_cs)

                prod_ns.append(ProdNodes(ns.num_node_blocks, chs, edge_ids, block_size = ns.block_size))

            if len(prod_ns) == 1:
                new_ns = prod_ns[0]
            else:
                new_ns = merge_prod_nodes(*prod_ns)
            sid = 0
            for ns in rg2nodes[rg_hash]:
                nid = sid + ns.num_node_blocks
                ns_old2new[ns] = (new_ns, (sid, nid))
                sid = nid

        elif isinstance(rg, InnerRegionNode):
            sum_ns = []
            for ns in rg2nodes[rg_hash]:
                chs = []
                ch2sid = dict()
                edge_ids = ns.edge_ids.clone()
                global_sid = 0
                origin_sid = 0
                for scope_id, cs in enumerate(ns.chs):
                    origin_eid = origin_sid + cs.num_node_blocks
                    new_cs, (offset_sid, offset_eid) = ns_old2new[cs]
                    if new_cs in ch2sid:
                        sid = ch2sid[new_cs]
                    else:
                        sid = global_sid

                    filter = (ns.edge_ids[1,:] >= origin_sid) & (ns.edge_ids[1,:] < origin_eid)
                    edge_ids[1,filter] += sid + offset_sid - origin_sid

                    if new_cs not in ch2sid:
                        chs.append(new_cs)
                        ch2sid[new_cs] = global_sid
                        global_sid += new_cs.num_node_blocks

                    origin_sid = origin_eid

                if hasattr(ns, "_params") and ns._params is not None:
                    params = ns._params
                else:
                    params = None
                sum_ns.append(SumNodes(ns.num_node_blocks, chs, edge_ids, params = params, block_size = ns.block_size))

            if len(sum_ns) == 1:
                new_ns = sum_ns[0]
            else:
                new_ns = merge_sum_nodes(*sum_ns)
            sid = 0
            for ns in rg2nodes[rg_hash]:
                nid = sid + ns.num_node_blocks
                ns_old2new[ns] = (new_ns, (sid, nid))
                sid = nid

    return ns_old2new[root_ns][0]


def merge(ns1: CircuitNodes, *args) -> CircuitNodes:
    """
    Merge nodes with identical region node together.

    :param ns1: the first PC node
    :type ns1: CircuitNodes

    :param args: the remaining PC nodes
    :type args: CircuitNodes

    Example::
        >>> i00 = inputs(0, num_node_blocks, dists.Categorical(num_cats = 5))
        >>> i01 = inputs(0, num_node_blocks, dists.Categorical(num_cats = 5))
        >>> i10 = inputs(1, num_node_blocks, dists.Categorical(num_cats = 5))
        >>> i11 = inputs(1, num_node_blocks, dists.Categorical(num_cats = 5))
 
        >>> m00 = multiply(i00, i10)
        >>> m01 = multiply(i01, i11)

        >>> n0 = summate(m00, num_node_blocks = num_node_blocks)
        >>> n1 = summate(m01, num_node_blocks = num_node_blocks)

        >>> n_new = pyjuice.merge(n0, n1)
    """
    if ns1.is_sum() and len(args) > 0 and args[0].is_sum():
        return merge_sum_nodes(ns1, args[0], *args[1:])
    elif ns1.is_prod() and len(args) > 0 and args[0].is_prod():
        return merge_prod_nodes(ns1, args[0], *args[1:])
    elif len(args) == 0:
        return merge_by_region_node(ns1)
    else:
        raise ValueError()


from __future__ import annotations

import numpy as np
import torch

from typing import Optional, Dict, Union, Sequence, Tuple

from pyjuice.nodes import CircuitNodes, InputNodes, ProdNodes, SumNodes, foreach, foldup_aggregate, summate, multiply

Tensor = Union[np.ndarray,torch.Tensor]


def prune_by_score(root_nodes: CircuitNodes, key: str = "_scores", scores: Optional[Dict[CircuitNodes,Tensor]] = None, 
                   keep_frac: Optional[float] = None, score_threshold: Optional[float] = None, block_reduction: str = "sum"):
    
    # Traverse all nodes to collect scores
    score_ranges = dict()
    score_start = 0
    flat_scores = []

    def _get_scores(ns: CircuitNodes):

        nonlocal score_start

        if scores is not None:
            if ns in scores:
                curr_scores = scores[ns]
            else:
                curr_scores = None
        else:
            if hasattr(ns, key):
                curr_scores = ns.__dict__[key]
            else:
                curr_scores = None

        if curr_scores is None or ns._source_node is not None:
            # We do not consider nodes with no specified scores and parameter-tied nodes
            return None

        if isinstance(curr_scores, np.ndarray):
            curr_scores = torch.from_numpy(curr_scores)

        if curr_scores.dim() == 3:
            if block_reduction == "sum":
                curr_scores = curr_scores.sum(dim = 2).sum(dim = 1)
            elif block_reduction == "mean":
                curr_scores = curr_scores.mean(dim = 2).mean(dim = 1)
            elif block_reduction == "max":
                curr_scores = curr_scores.max(dim = 2).max(dim = 1)
            else:
                raise ValueError(f"Unknown block reduction method `{block_reduction}`.")
        else:
            assert curr_scores.dim() == 1

        flat_scores.append(curr_scores)
        num_scores = curr_scores.size(0)

        score_ranges[ns] = (score_start, score_start + num_scores)
        score_start += num_scores

    for ns in root_nodes:
        _get_scores(ns)

    # Indices to keep
    flat_scores = torch.cat(flat_scores, dim = 0)
    if flat_scores.dim() == 3:
        flat_scores = flat_scores.sum(dim = 2).sum(dim = 1)
    if keep_frac is not None:
        assert score_threshold is None, "Only one of `keep_frac` and `score_threshold` should be set."
        score_threshold = torch.quantile(flat_scores, 1.0 - keep_frac, dim = 0)
        selected_edges = (flat_scores >= score_threshold)
    else:
        assert score_threshold is not None, "At least one of `keep_frac` and `score_threshold` should be set."
        selected_edges = (flat_scores >= score_threshold)

    # Prune circuit

    def _construct_pruned_circuit(ns: CircuitNodes, ch_outputs: Sequence[CircuitNodes], dup2source: Dict[CircuitNodes,CircuitNodes] = dict()):
        if isinstance(ns, InputNodes):
            new_ns = ns.duplicate(tie_params = False)

        elif isinstance(ns, ProdNodes):
            new_ns = ns.duplicate(*ch_outputs, tie_params = False)

        else:
            assert isinstance(ns, SumNodes)

            if ns in score_ranges:
                # Prune away edge parameters
                edge_ids = ns.edge_ids.clone()
                edge_filter = selected_edges[score_ranges[ns][0]:score_ranges[ns][1]]
                copied_edges = []
                copied_params = []
                for node_id in range(ns.num_node_blocks):
                    curr_eids = (edge_ids[0,:] == node_id) * edge_filter
                    if curr_eids.sum().item() == 0:
                        maxid = torch.argmax(
                            flat_scores[score_ranges[ns][0]:score_ranges[ns][1]] * (edge_ids[0,:] == node_id) + 
                            (edge_ids[0,:] == node_id) * 1e-8
                        )
                        copied_edges.append(edge_ids[:,maxid].unsqueeze(1))
                        copied_params.append(ns._params[maxid].reshape(1, ns.block_size, ns.ch_block_size))
                    else:
                        copied_edges.append(edge_ids[:,curr_eids])
                        copied_params.append(ns._params[curr_eids,:,:])

                edge_ids = torch.cat(copied_edges, dim = 1)
                params = torch.cat(copied_params, dim = 0)

                assert edge_ids.size(1) == params.size(0)

                new_ns = SumNodes(
                    num_node_blocks = ns.num_node_blocks, 
                    chs = ch_outputs, 
                    edge_ids = edge_ids,
                    params = params,
                    block_size = ns.block_size
                )
            else:
                # Keep the node as-is
                new_ns = ns.duplicate(*ch_outputs, tie_params = False)

        if ns._source_node is not None:
            dup2source[new_ns] = ns._source_node

        return new_ns

    old2new = dict()
    dup2source = dict()
    new_root_ns = foldup_aggregate(_construct_pruned_circuit, root_nodes, cache = old2new, dup2source = dup2source)

    for ns in dup2source:
        ns_source = old2new[dup2source[ns]]
        ns._source_node = ns_source

        assert ns.num_node_blocks == ns_source.num_node_blocks and ns.block_size == ns_source.block_size

        if hasattr(ns_source, "edge_ids"):
            ns.edge_ids = ns_source.edge_ids.clone()

    return new_root_ns


from __future__ import annotations

import numpy as np
import torch
import triton
import triton.language as tl
from copy import deepcopy as pydeepcopy
from numba import njit
from typing import Optional, Dict, Sequence

from pyjuice.nodes import CircuitNodes, InputNodes, ProdNodes, SumNodes, foldup_aggregate
from pyjuice.utils import BitSet
from pyjuice.utils.util import max_cdf_power_of_2
from pyjuice.utils.kernel_launcher import FastJITFunction


@njit
def _compute_param_target_ids_kernel(target_id0, target_id1, target_id2, edge_ids, new_edge_ids, 
                                     block_mul_size, ch_block_mul_size, block_size, ch_block_size):
    for i in range(edge_ids.shape[1]):
        old_ngid = edge_ids[0,i]
        old_cgid = edge_ids[1,i]

        for j in range(new_edge_ids.shape[1]):
            new_ngid = new_edge_ids[0,j]
            new_cgid = new_edge_ids[1,j]
            
            ng_sid = new_ngid * block_mul_size
            ng_eid = ng_sid + block_mul_size
            cg_sid = new_cgid * ch_block_mul_size
            cg_eid = cg_sid + ch_block_mul_size

            if (old_ngid >= ng_sid) and (old_ngid < ng_eid) and (old_cgid >= cg_sid) and (old_cgid < cg_eid):
                target_id0[i] = j

                target_id1[0,i] = (old_ngid - ng_sid) * block_size
                target_id1[1,i] = (old_ngid - ng_sid + 1) * block_size

                target_id2[0,i] = (old_cgid - cg_sid) * ch_block_size
                target_id2[1,i] = (old_cgid - cg_sid + 1) * ch_block_size
                break


@triton.jit
def _copy_params_kernel(new_params, mparams, target_id0, target_id1, target_id2, 
                        old_block_size: tl.constexpr, old_ch_block_size: tl.constexpr, 
                        new_block_size: tl.constexpr, new_ch_block_size: tl.constexpr, 
                        BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr):

    pid_n = tl.program_id(0)
    pid_m = tl.program_id(1)
    pid_b = tl.program_id(2)

    offs_m = tl.arange(0, BLOCK_M) + pid_m * BLOCK_M
    offs_n = tl.arange(0, BLOCK_N) + pid_n * BLOCK_N

    offs_pars = pid_b * (old_block_size * old_ch_block_size) + offs_m[:,None] * old_ch_block_size + offs_n[None,:]
    pars = tl.load(mparams + offs_pars)

    id0 = tl.load(target_id0 + pid_b)
    id1 = tl.load(target_id1 + pid_b)
    id2 = tl.load(target_id2 + pid_b)

    offs_npars = id0 * (new_block_size * new_ch_block_size) + (id1 + offs_m)[:,None] * new_ch_block_size + (id2 + offs_n)[None,:]
    tl.store(new_params + offs_npars, pars)


def blockify(root_ns: CircuitNodes, sparsity_tolerance: float = 0.25, max_target_block_size: int = 32, use_cuda: bool = True) -> CircuitNodes:
    """
    Generate an equivalent PC with potentially high block sizes.

    :param root_ns: the input PC
    :type root_ns: CircuitNodes

    :param sparsity_tolerance: allowed fraction of zero parameters to be added (should be in the range (0, 1])
    :type sparsity_tolerance: float

    :param max_target_block_size: the maximum block size to search for
    :type max_target_block_size: int

    :param use_cuda: use GPU when possible
    :type use_cuda: bool

    :returns: An equivalent `CircuitNodes`
    """

    if use_cuda:
        device = torch.device("cuda:0")
    else:
        device = torch.device("cpu")

    ## Do an initial pass to compute the maximum block size of every `ns` ##

    ns2block_size = dict()
    for ns in root_ns:
        if ns.is_input():
            ns2block_size[ns] = min(max_cdf_power_of_2(ns.num_nodes), max(ns.block_size, max_target_block_size))

        elif ns.is_prod():
            ns2block_size[ns] = min(max_cdf_power_of_2(ns.num_nodes), max(ns.block_size, max_target_block_size))

        else:
            assert ns.is_sum()

            old_block_size = ns.block_size
            old_ch_block_size = ns.ch_block_size
            edge_ids = ns.edge_ids

            old_ns_num_nblocks = ns.num_node_blocks
            old_cs_num_nblocks = sum([cs.num_node_blocks for cs in ns.chs])

            flag = False
            plausible_combinations = list()

            block_size = min(max_cdf_power_of_2(ns.num_nodes), max(ns.block_size, max_target_block_size))
            while block_size >= old_block_size:
                block_mul_size = block_size // old_block_size

                ns_num_nblocks = old_ns_num_nblocks // block_mul_size
                
                ch_block_size = ns2block_size[ns.chs[0]]
                while ch_block_size >= old_ch_block_size:
                    ch_block_mul_size = ch_block_size // old_ch_block_size

                    cs_num_nblocks = old_cs_num_nblocks // block_mul_size

                    n_edge_ids = edge_ids[0,:] // block_mul_size
                    c_edge_ids = edge_ids[1,:] // ch_block_mul_size
                    _, counts = torch.unique(n_edge_ids * cs_num_nblocks + c_edge_ids, return_counts = True)

                    if counts.float().mean() >= (1.0 - sparsity_tolerance) * block_mul_size * ch_block_mul_size:
                        plausible_combinations.append((block_size, ch_block_size))

                    ch_block_size = ch_block_size // 2

                block_size = block_size // 2

            # Find the best block size combination
            best_block_size = 0
            best_ch_block_size = 0
            for block_size, ch_block_size in plausible_combinations:
                if block_size >= 16 and ch_block_size >= 16:
                    best_block_size = block_size
                    best_ch_block_size = ch_block_size
                    break

            if best_block_size == 0:
                best_val = 0
                best_frac = 0
                for block_size, ch_block_size in plausible_combinations:
                    cond1 = block_size * ch_block_size > best_val
                    cond2 = (block_size * ch_block_size > best_val) and \
                        (max(block_size, ch_block_size) // min(block_size, ch_block_size) < best_frac)
                    if cond1 or cond2:
                        best_block_size = block_size
                        best_ch_block_size = ch_block_size
                        best_val = block_size * ch_block_size
                        best_frac = max(block_size, ch_block_size) // min(block_size, ch_block_size)

            if best_block_size == 0:
                best_block_size = old_block_size
                best_ch_block_size = old_ch_block_size

            ns2block_size[ns] = best_block_size
            for cs in ns.chs:
                ns2block_size[cs] = best_ch_block_size

    ## Do a second pass to finalize the block sizes ##

    for ns in root_ns:
        if ns.is_prod():
            block_size = ns2block_size[ns]
            for cs in ns.chs:
                block_size = min(block_size, ns2block_size[cs])

            ns2block_size[ns] = block_size
            for cs in ns.chs:
                ns2block_size[cs] = block_size

    ## Apply the new block sizes ##

    def update_ns(ns: CircuitNodes, ns_chs: Sequence[CircuitNodes]):
        new_block_size = ns2block_size[ns]
        block_mul_size = new_block_size // ns.block_size

        new_num_nblocks = ns.num_node_blocks // block_mul_size

        assert new_num_nblocks * new_block_size == ns.num_node_blocks * ns.block_size

        if ns.is_input():
            new_ns = InputNodes(
                num_node_blocks = new_num_nblocks,
                scope = pydeepcopy(ns.scope),
                dist = pydeepcopy(ns.dist),
                block_size = new_block_size
            )

            if not ns.is_tied():
                params = ns.get_params()
                if params is not None:
                    new_ns.set_params(params.clone(), normalize = False)

        elif ns.is_prod():
            edge_ids = ns.edge_ids.clone()
            edge_ids = edge_ids.reshape(new_num_nblocks, block_mul_size, ns.num_chs)
            if torch.all(edge_ids[:,1:,:] - edge_ids[:,:-1,:]) == 1:
                # Block-sparse mode
                edge_ids = edge_ids[:,0,:].contiguous() // block_mul_size
                mode = "block_sparse"
            else:
                # Sparse mode
                edge_ids = (edge_ids.reshape(ns.num_node_blocks, ns.num_chs)[:,None,:] * ns.block_size + \
                    torch.arange(0, ns.block_size)[None,:,None]).flatten(0, 1)
                mode = "sparse"

            new_ns = ProdNodes(
                num_node_blocks = new_num_nblocks,
                chs = ns_chs,
                edge_ids = edge_ids,
                block_size = new_block_size
            )

            if mode == "block_sparse":
                assert new_ns.is_block_sparse()
            elif mode == "sparse":
                assert new_ns.is_sparse()

        else:
            assert ns.is_sum()

            old_num_nblocks = ns.num_node_blocks
            old_num_cblocks = sum([cs.num_node_blocks for cs in ns.chs])

            new_ch_block_size = ns2block_size[ns.chs[0]]
            ch_block_mul_size = new_ch_block_size // ns.chs[0].block_size

            new_num_cblocks = old_num_cblocks // ch_block_mul_size

            edge_ids = ns.edge_ids.clone()
            grid_edge_ids = torch.zeros([old_num_nblocks, old_num_cblocks], dtype = torch.bool)
            grid_edge_ids[edge_ids[0,:],edge_ids[1,:]] = True

            grid_edge_ids = grid_edge_ids.reshape(new_num_nblocks, block_mul_size, new_num_cblocks, ch_block_mul_size)
            new_edge_ids = torch.nonzero(grid_edge_ids.any(dim = 3).any(dim = 1), as_tuple = False).permute(1, 0)

            new_ns = SumNodes(
                num_node_blocks = new_num_nblocks,
                chs = ns_chs,
                edge_ids = new_edge_ids,
                block_size = new_block_size
            )
            
            if not ns.is_tied():
                # Collect selected blocks
                grid_edge_ids = grid_edge_ids.permute(0, 2, 1, 3).flatten(0, 1)
                block_ids = new_edge_ids[0,:] * new_num_cblocks + new_edge_ids[1,:]
                param_indicator = grid_edge_ids[block_ids,:,:]
                if not torch.all(param_indicator):
                    param_indicator = param_indicator[:,:,None,:,None].repeat(1, 1, ns.block_size, 1, ns.chs[0].block_size)
                    param_indicator = param_indicator.flatten(3, 4).flatten(1, 2)
                    zero_param_mask = ~param_indicator

                    new_ns.set_zero_param_mask(zero_param_mask)

                params = ns.get_params()
                if params is not None:
                    old_block_size = ns.block_size
                    old_ch_block_size = ns.chs[0].block_size
                    if new_block_size > old_block_size or new_ch_block_size > old_ch_block_size:
                        new_params = torch.zeros([new_edge_ids.size(1), new_block_size, new_ch_block_size], device = device)
                        if new_edge_ids.size(1) == new_num_nblocks * new_num_cblocks and params.numel() == new_params.numel():
                            # Fully-connected parameters
                            new_params = params.reshape(
                                new_num_nblocks, block_mul_size, new_num_cblocks, ch_block_mul_size, old_block_size, old_ch_block_size
                            ).permute(0, 2, 1, 4, 3, 5).reshape(new_params.size()).contiguous()
                        elif use_cuda:
                            edge_ids_np = edge_ids.numpy()
                            new_edge_ids_np = new_edge_ids.numpy()

                            target_id0 = np.zeros([edge_ids.size(1)], dtype = np.int64) - 1
                            target_id1 = np.zeros([2, edge_ids.size(1)], dtype = np.int64) - 1
                            target_id2 = np.zeros([2, edge_ids.size(1)], dtype = np.int64) - 1
                            
                            _compute_param_target_ids_kernel(
                                target_id0, target_id1, target_id2, edge_ids_np, new_edge_ids_np, 
                                block_mul_size, ch_block_mul_size, old_block_size, old_ch_block_size
                            )

                            target_id0 = torch.from_numpy(target_id0).to(device)
                            target_id1 = torch.from_numpy(target_id1).to(device)
                            target_id2 = torch.from_numpy(target_id2).to(device)

                            params = params.to(device)

                            BLOCK_M = min(32, old_block_size)
                            BLOCK_N = min(32, old_ch_block_size)

                            grid = (old_ch_block_size // BLOCK_N, old_block_size // BLOCK_M, edge_ids.size(1))

                            _copy_params_kernel[grid](
                                new_params, params, target_id0, target_id1, target_id2, 
                                old_block_size = old_block_size, 
                                old_ch_block_size = old_ch_block_size, 
                                new_block_size = new_block_size, 
                                new_ch_block_size = new_ch_block_size, 
                                BLOCK_M = BLOCK_M, 
                                BLOCK_N = BLOCK_N
                            )

                        else:
                            for par_block_id in range(new_edge_ids.size(1)):
                                nsid = new_edge_ids[0,par_block_id] * block_mul_size
                                neid = nsid + block_mul_size
                                csid = new_edge_ids[1,par_block_id] * ch_block_mul_size
                                ceid = csid + ch_block_mul_size

                                blk_ids = torch.where((edge_ids[0,:] >= nsid) & (edge_ids[0,:] < neid) & (edge_ids[1,:] >= csid) & (edge_ids[1,:] < ceid))[0]
                                for blk_id in blk_ids:
                                    nid0, nid1 = (edge_ids[0,blk_id] - nsid) * ns.block_size, (edge_ids[0,blk_id] - nsid + 1) * ns.block_size
                                    cid0, cid1 = (edge_ids[1,blk_id] - csid) * ns.chs[0].block_size, (edge_ids[1,blk_id] - csid + 1) * ns.chs[0].block_size
                                    new_params[par_block_id,nid0:nid1,cid0:cid1] = params[blk_id,:,:]

                        new_ns.set_params(new_params.cpu(), normalize = False)
                    else:
                        new_ns.set_params(params, normalize = False)

        return new_ns

    old2new = dict()
    new_root_ns = foldup_aggregate(update_ns, root_ns, cache = old2new)

    # Re-link tied nodes to their source
    for ns in root_ns:
        if ns.is_tied():
            new_source_ns = old2new[ns.get_source_ns()]
            new_ns = old2new[ns]
            new_ns.set_source_ns(new_source_ns)

    return new_root_ns


def unblockify(root_ns: CircuitNodes, block_size: int = 1, recursive: bool = True, keys_to_copy: Optional[Sequence[str]] = None):
    """
    Decrease the block size of a PC.

    :param root_ns: the input PC
    :type root_ns: CircuitNodes

    :param block_size: the target block size
    :type block_size: int

    :param recursive: whether to do it recursively or just for the current node
    :type recursive: bool

    :param keys_to_copy: an optional dictionary of properties to copy
    :type keys_to_copy: Optional[Sequence[str]]

    :returns: An equivalent `CircuitNodes`
    """
    
    def update_ns(ns: CircuitNodes, ns_chs: Sequence[CircuitNodes]):
        new_block_size = min(block_size, ns.block_size)
        new_num_nblocks = ns.num_nodes // new_block_size

        if ns.is_input():
            new_ns = InputNodes(
                num_node_blocks = new_num_nblocks,
                scope = pydeepcopy(ns.scope),
                dist = pydeepcopy(ns.dist),
                block_size = new_block_size
            )

            if not ns.is_tied():
                params = ns.get_params()
                if params is not None:
                    new_ns.set_params(params.clone(), normalize = False)

        elif ns.is_prod():
            if ns.is_block_sparse():
                gsize_redu = ns.block_size // new_block_size
                edge_ids = ns.edge_ids.clone()
                edge_ids = (edge_ids[:,None,:].repeat(1, gsize_redu, 1) * gsize_redu + \
                    torch.arange(0, gsize_redu)[None,:,None]).reshape(ns.num_node_blocks * gsize_redu, ns.num_chs)

            else:
                edge_ids = ns.edge_ids.clone()

            new_ns = ProdNodes(
                num_node_blocks = new_num_nblocks,
                chs = ns_chs,
                edge_ids = edge_ids,
                block_size = new_block_size
            )

        else:
            assert ns.is_sum()

            new_ch_block_size = ns_chs[0].block_size
            gsize_redu = ns.block_size // new_block_size
            ch_gsize_redu = ns.ch_block_size // new_ch_block_size

            grid_x, grid_y = torch.meshgrid(
                torch.arange(0, gsize_redu), 
                torch.arange(0, ch_gsize_redu), 
                indexing = 'ij'
            )

            edge_ids = ns.edge_ids.clone()
            edge_ids = edge_ids[:,None,:].repeat(1, gsize_redu * ch_gsize_redu, 1)
            edge_ids[0,:,:] = edge_ids[0,:,:] * gsize_redu + grid_x.reshape(-1)[:,None]
            edge_ids[1,:,:] = edge_ids[1,:,:] * ch_gsize_redu + grid_y.reshape(-1)[:,None]
            edge_ids = edge_ids.reshape(2, ns.edge_ids.size(1) * gsize_redu * ch_gsize_redu)

            new_ns = SumNodes(
                num_node_blocks = new_num_nblocks,
                chs = ns_chs,
                edge_ids = edge_ids,
                block_size = new_block_size
            )
            
            if not ns.is_tied() and ns.has_params():
                params = ns._params.clone()
                params = params.reshape(
                    ns.edge_ids.size(1), gsize_redu, new_block_size, ch_gsize_redu, new_ch_block_size
                ).permute(0, 1, 3, 2, 4).reshape(
                    ns.edge_ids.size(1) * gsize_redu * ch_gsize_redu, new_block_size, new_ch_block_size
                )

                new_ns.set_params(params)

                if keys_to_copy is not None:
                    for key in keys_to_copy:
                        if hasattr(ns, key):
                            vals = getattr(ns, key)

                            if vals.dim() == 3:
                                vals = vals.reshape(
                                    ns.edge_ids.size(1), gsize_redu, new_block_size, ch_gsize_redu, new_ch_block_size
                                ).permute(0, 1, 3, 2, 4).reshape(
                                    ns.edge_ids.size(1) * gsize_redu * ch_gsize_redu, new_block_size, new_ch_block_size
                                )
                                setattr(new_ns, key, vals)
                            else:
                                raise NotImplementedError()

                assert new_ns._params.size(0) == new_ns.edge_ids.size(1)

        return new_ns

    old2new = dict()
    new_root_ns = foldup_aggregate(update_ns, root_ns, cache = old2new)

    # Re-link tied nodes to their source
    for ns in root_ns:
        if ns.is_tied():
            new_source_ns = old2new[ns.get_source_ns()]
            new_ns = old2new[ns]
            new_ns.set_source_ns(new_source_ns)

    return new_root_ns


def bump_block_size(ns: CircuitNodes, block_size: int, use_cuda: bool = True):
    assert block_size > ns.block_size, f"`block_size` already greater than {block_size}."
    assert ns.num_nodes % block_size == 0, f"`num_nodes` not divicible by the target block size."

    if use_cuda:
        device = torch.device("cuda:0")
    else:
        device = torch.device("cpu")

    block_mul_size = block_size // ns.block_size

    if ns.is_input():
        new_ns = InputNodes(
            num_node_blocks = new_num_nblocks,
            scope = pydeepcopy(ns.scope),
            dist = pydeepcopy(ns.dist),
            block_size = block_size
        )

        if not ns.is_tied():
            params = ns.get_params()
            if params is not None:
                new_ns.set_params(params.clone(), normalize = False)

    elif ns.is_prod():
        edge_ids = ns.edge_ids.clone()
        edge_ids = edge_ids.reshape(new_num_nblocks, block_mul_size, ns.num_chs)
        if torch.all(edge_ids[:,1:,:] - edge_ids[:,:-1,:]) == 1:
            # Block-sparse mode
            edge_ids = edge_ids[:,0,:].contiguous()
            mode = "block_sparse"
        else:
            # Sparse mode
            edge_ids = (edge_ids.reshape(ns.num_node_blocks, ns.num_chs)[:,None,:] * ns.block_size + \
                torch.arange(0, ns.block_size)[None,:,None]).flatten(0, 1)
            mode = "sparse"
        
        new_ns = ProdNodes(
            num_node_blocks = new_num_nblocks,
            chs = ns_chs,
            edge_ids = edge_ids,
            block_size = block_size
        )

        if mode == "block_sparse":
            assert new_ns.is_block_sparse()
        elif mode == "sparse":
            assert new_ns.is_sparse()
    if ns.is_sum():
        old_num_nblocks = ns.num_node_blocks
        old_block_size = ns.block_size
        num_cblocks = sum([cs.num_node_blocks for cs in ns.chs])

        new_num_nblocks = ns.num_nodes // block_size

        ns_chs = ns.chs
        ch_block_size = ns.chs[0].block_size

        edge_ids = ns.edge_ids.clone()
        grid_edge_ids = torch.zeros([old_num_nblocks, num_cblocks], dtype = torch.bool)
        grid_edge_ids[edge_ids[0,:],edge_ids[1,:]] = True

        grid_edge_ids = grid_edge_ids.reshape(new_num_nblocks, block_mul_size, num_cblocks, 1)
        new_edge_ids = torch.nonzero(grid_edge_ids.any(dim = 3).any(dim = 1), as_tuple = False).permute(1, 0)

        new_ns = SumNodes(
            num_node_blocks = new_num_nblocks,
            chs = ns_chs,
            edge_ids = new_edge_ids,
            block_size = block_size
        )

        if not ns.is_tied():
            # Collect selected blocks
            grid_edge_ids = grid_edge_ids.permute(0, 2, 1, 3).flatten(0, 1)
            block_ids = new_edge_ids[0,:] * num_cblocks + new_edge_ids[1,:]
            param_indicator = grid_edge_ids[block_ids,:,:]
            if not torch.all(param_indicator):
                param_indicator = param_indicator[:,:,None,:,None].repeat(1, 1, ns.block_size, 1, ns.chs[0].block_size)
                param_indicator = param_indicator.flatten(3, 4).flatten(1, 2)
                zero_param_mask = ~param_indicator

                new_ns.set_zero_param_mask(zero_param_mask)

            params = ns.get_params()
            if params is not None:
                new_params = torch.zeros([new_edge_ids.size(1), block_size, ch_block_size], device = device)
                if new_edge_ids.size(1) == new_num_nblocks * num_cblocks:
                    # Fully-connected parameters
                    new_params = params.reshape(
                        new_num_nblocks, block_size // old_block_size, num_cblocks, 1, old_block_size, ch_block_size
                    ).permute(0, 2, 1, 4, 3, 5).reshape(new_params.size()).contiguous()
                elif use_cuda:
                    edge_ids_np = edge_ids.numpy()
                    new_edge_ids_np = new_edge_ids.numpy()

                    target_id0 = np.zeros([edge_ids.size(1)], dtype = np.int64) - 1
                    target_id1 = np.zeros([2, edge_ids.size(1)], dtype = np.int64) - 1
                    target_id2 = np.zeros([2, edge_ids.size(1)], dtype = np.int64) - 1
                    
                    _compute_param_target_ids_kernel(
                        target_id0, target_id1, target_id2, edge_ids_np, new_edge_ids_np, 
                        block_mul_size, 1, old_block_size, ch_block_size
                    )

                    target_id0 = torch.from_numpy(target_id0).to(device)
                    target_id1 = torch.from_numpy(target_id1).to(device)
                    target_id2 = torch.from_numpy(target_id2).to(device)

                    params = params.to(device)

                    BLOCK_M = min(32, old_block_size)
                    BLOCK_N = min(32, ch_block_size)

                    grid = (ch_block_size // BLOCK_N, old_block_size // BLOCK_M, edge_ids.size(1))

                    _copy_params_kernel[grid](
                        new_params, params, target_id0, target_id1, target_id2, 
                        old_block_size = old_block_size, 
                        old_ch_block_size = ch_block_size, 
                        new_block_size = block_size, 
                        new_ch_block_size = ch_block_size, 
                        BLOCK_M = BLOCK_M, 
                        BLOCK_N = BLOCK_N
                    )

                else:
                    for par_block_id in range(new_edge_ids.size(1)):
                        nsid = new_edge_ids[0,par_block_id] * block_mul_size
                        neid = nsid + block_mul_size
                        csid = new_edge_ids[1,par_block_id]
                        ceid = csid + 1

                        blk_ids = torch.where((edge_ids[0,:] >= nsid) & (edge_ids[0,:] < neid) & (edge_ids[1,:] >= csid) & (edge_ids[1,:] < ceid))[0]
                        for blk_id in blk_ids:
                            nid0, nid1 = (edge_ids[0,blk_id] - nsid) * ns.block_size, (edge_ids[0,blk_id] - nsid + 1) * ns.block_size
                            cid0, cid1 = (edge_ids[1,blk_id] - csid) * ns.chs[0].block_size, (edge_ids[1,blk_id] - csid + 1) * ns.chs[0].block_size
                            new_params[par_block_id,nid0:nid1,cid0:cid1] = params[blk_id,:,:]

                new_ns.set_params(new_params.cpu(), normalize = False)

    return new_ns
